"""Unit tests for the problem_report module."""

# pylint: disable=too-many-lines

import base64
import contextlib
import datetime
import email
import io
import json
import locale
import sys
import textwrap
import time
import unittest
import unittest.mock
from unittest.mock import MagicMock, patch

try:
    import zstandard
except ImportError:
    zstandard = None  # type: ignore

import problem_report

BIN_DATA = b"ABABABABAB\0\0\0Z\x01\x02"
GZIP_BIN_DATA = (
    b"\x1f\x8b\x08\x00\x00\x00\x00\x00\x00\x03str\x84B\x06"
    b"\x06\x86(F&\x003\x95\xd4\x0b\x10\x00\x00\x00"
)
ZSTD_BIN_DATA = (
    b"(\xb5/\xfd$\x10\x81\x00\x00ABABABABAB\x00\x00\x00Z\x01\x02\xbc\xdf\xdd\xfd"
)


class T(unittest.TestCase):  # pylint: disable=too-many-public-methods
    """Unit tests for the problem_report module."""

    def test_add_tags(self) -> None:
        """Test ProblemReport.add_tags()."""
        report = problem_report.ProblemReport()
        report.add_tags({"tag1"})
        self.assertEqual(report["Tags"], "tag1")
        report.add_tags(["tag2", "next_tag"])
        self.assertEqual(report["Tags"], "next_tag tag1 tag2")

    def test_add_tag_drop_duplicates(self) -> None:
        """Test ProblemReport.add_tags() dropping duplicates."""
        report = problem_report.ProblemReport()
        report.add_tags({"same"})
        self.assertEqual(report["Tags"], "same")
        report.add_tags(["same"])
        self.assertEqual(report["Tags"], "same")

    def test_basic_operations(self) -> None:
        """Basic creation and operation."""
        pr = problem_report.ProblemReport()
        pr["foo"] = "bar"
        pr["bar"] = " foo   bar\nbaz\n   blip  "
        pr["dash-key"] = "1"
        pr["dot.key"] = "1"
        pr["underscore_key"] = "1"
        self.assertEqual(pr["foo"], "bar")
        self.assertEqual(pr["bar"], " foo   bar\nbaz\n   blip  ")
        self.assertEqual(pr["ProblemType"], "Crash")
        locale_time = locale.getlocale(locale.LC_TIME)
        locale.setlocale(locale.LC_TIME, "C")
        try:
            self.assertTrue(time.strptime(pr["Date"]))
        finally:
            locale.setlocale(locale.LC_TIME, locale_time)
        self.assertEqual(pr["dash-key"], "1")
        self.assertEqual(pr["dot.key"], "1")
        self.assertEqual(pr["underscore_key"], "1")

    def test_ctor_arguments(self) -> None:
        """non-default constructor arguments."""
        pr = problem_report.ProblemReport("KernelCrash")
        self.assertEqual(pr["ProblemType"], "KernelCrash")
        pr = problem_report.ProblemReport(date="19801224 12:34")
        self.assertEqual(pr["Date"], "19801224 12:34")

    def test_get_timestamp(self) -> None:
        """get_timestamp() returns timestamp."""
        r = problem_report.ProblemReport()
        timestamp = r.get_timestamp()
        assert timestamp
        self.assertAlmostEqual(timestamp, time.time(), delta=2)

        r["Date"] = "Thu Jan 9 12:00:00 2014"
        # delta is ±12 hours, as this depends on the timezone that the test is
        # run in
        timestamp = r.get_timestamp()
        assert timestamp
        self.assertAlmostEqual(timestamp, 1389265200, delta=43200)

    def test_get_timestamp_locale_german(self) -> None:
        """get_timestamp() returns date when LC_TIME is set."""
        now = datetime.datetime.now()

        pr = problem_report.ProblemReport(date=now.strftime("%a %b %d %H:%M:%S %Y"))
        orig_ctime = locale.getlocale(locale.LC_TIME)
        try:
            locale.setlocale(locale.LC_TIME, "de_DE.UTF-8")
        except locale.Error:
            self.skipTest("Missing German locale support")
        self.assertEqual(pr.get_timestamp(), int(now.timestamp()))
        locale.setlocale(locale.LC_TIME, orig_ctime)

    def test_get_timestamp_returns_none(self) -> None:
        """get_timestamp() returns None."""
        pr = problem_report.ProblemReport()
        del pr["Date"]
        self.assertIsNone(pr.get_timestamp())

    def test_consistency_checks(self) -> None:
        """Various error conditions."""
        pr = problem_report.ProblemReport()
        self.assertRaises(
            problem_report.InvalidProblemReportKeyError, pr.__setitem__, "a b", "1"
        )
        self.assertRaises(TypeError, pr.__setitem__, "a", 1)
        self.assertRaises(TypeError, pr.__setitem__, "a", (1,))
        self.assertRaises(TypeError, pr.__setitem__, "a", ("/tmp/nonexistent", ""))
        self.assertRaises(
            TypeError,
            pr.__setitem__,
            "a",
            ("/tmp/nonexistent", False, 0, True, "bogus"),
        )
        self.assertRaises(TypeError, pr.__setitem__, "a", ["/tmp/nonexistent"])
        self.assertRaises(KeyError, pr.__getitem__, "Nonexistent")

    def test_write(self) -> None:
        """write() and proper formatting."""
        pr = problem_report.ProblemReport(date="now!")
        pr["Simple"] = "bar"
        pr["SimpleUTF8"] = "1äö2Φ3".encode()
        pr["SimpleUnicode"] = "1äö2Φ3"
        pr["TwoLineUnicode"] = "pi-π\nnu-η"
        pr["TwoLineUTF8"] = "pi-π\nnu-η".encode()
        pr["WhiteSpace"] = " foo   bar\nbaz\n  blip  \n\nafteremptyline"
        # Unicode with a non-space low ASCII character \x05 in it
        pr["UnprintableUnicode"] = b"a\xc3\xa4\x05z1\xc3\xa9".decode("UTF-8")
        out = io.BytesIO()
        pr.write(out)
        expected = (textwrap.dedent("""\
            ProblemType: Crash
            Date: now!
            Simple: bar
            SimpleUTF8: 1äö2Φ3
            SimpleUnicode: 1äö2Φ3
            TwoLineUTF8:
             pi-π
             nu-η
            TwoLineUnicode:
             pi-π
             nu-η
            UnprintableUnicode: aä\x05z1é
            WhiteSpace:
            """) + "  foo   bar\n baz\n   blip  \n \n afteremptyline\n").encode("UTF-8")
        self.assertEqual(out.getvalue(), expected)

    def test_load(self) -> None:
        """load() with various formatting."""
        report = textwrap.dedent(f"""\
            ProblemType: Crash
            Date: now!
            Simple: bar
            WhiteSpace:
              foo   bar
             baz
               blip{'  '}
            """)
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report.encode()))
        self.assertEqual(pr["ProblemType"], "Crash")
        self.assertEqual(pr["Date"], "now!")
        self.assertEqual(pr["Simple"], "bar")
        self.assertEqual(pr["WhiteSpace"], " foo   bar\nbaz\n  blip  ")

        # test last field a bit more
        report += " \n"
        pr.load(io.BytesIO(report.encode()))
        self.assertEqual(pr["ProblemType"], "Crash")
        self.assertEqual(pr["Date"], "now!")
        self.assertEqual(pr["Simple"], "bar")
        self.assertEqual(pr["WhiteSpace"], " foo   bar\nbaz\n  blip  \n")

        # last field might not be \n terminated
        pr.load(io.BytesIO(textwrap.dedent("""\
            ProblemType: Crash
            Date: now!
            Simple: bar
            WhiteSpace:
             foo
             bar""").encode()))
        self.assertEqual(pr["ProblemType"], "Crash")
        self.assertEqual(pr["Date"], "now!")
        self.assertEqual(pr["Simple"], "bar")
        self.assertEqual(pr["WhiteSpace"], "foo\nbar")

        report = (
            "ProblemType: Crash\n"
            "WhiteSpace:\n"
            "  foo   bar\n"
            " baz\n"
            " \n"
            "   blip  \n"
            "Last: foo\n"
        )
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report.encode()))
        self.assertEqual(pr["WhiteSpace"], " foo   bar\nbaz\n\n  blip  ")
        self.assertEqual(pr["Last"], "foo")

        report += " \n"
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report.encode()))
        self.assertEqual(pr["WhiteSpace"], " foo   bar\nbaz\n\n  blip  ")
        self.assertEqual(pr["Last"], "foo\n")

        # empty lines in values must have a leading space in coding
        invalid_spacing = io.BytesIO(textwrap.dedent("""\
            WhiteSpace:
             first

             second
            """).encode())
        pr = problem_report.ProblemReport()
        self.assertRaises(ValueError, pr.load, invalid_spacing)

        # test that load() cleans up properly
        pr.load(io.BytesIO(b"ProblemType: Crash"))
        self.assertEqual(list(pr.keys()), ["ProblemType"])

    def test_load_string_with_null(self) -> None:
        """Test load() with a NULL character in a string key.

        Test case for https://launchpad.net/bugs/2146806
        """
        proc_maps = "7f8e5d61f000-7f8e5d620000 r--p 00005000 fc:00 190815U\0\0"
        content = f"ProblemType: Crash\n" f"Date: now!\n" f"ProcMaps: {proc_maps}\n"
        report = problem_report.ProblemReport()
        report.load(io.BytesIO(content.encode()))

        self.assertEqual(report["ProblemType"], "Crash")
        self.assertEqual(report["Date"], "now!")
        self.assertEqual(report["ProcMaps"], proc_maps)

    def test_load_binary_blob(self) -> None:
        """Throw exception when binary file (e.g. core) is loaded."""
        report = problem_report.ProblemReport()
        with (
            io.BytesIO(b"AB\xfc:CD") as report_file,
            self.assertRaisesRegex(
                problem_report.MalformedProblemReport,
                "codec can't decode byte 0xfc in position 2",
            ),
        ):
            report.load(report_file)

    def test_load_missing_colon(self) -> None:
        """Throw exception when key-value line misses a colon as separator."""
        report = problem_report.ProblemReport()
        with (
            io.BytesIO(b"\n") as report_file,
            self.assertRaisesRegex(
                problem_report.MalformedProblemReport,
                r"Line '\\n' does not contain a colon",
            ),
        ):
            report.load(report_file)

    def test_load_invalid_key(self) -> None:
        """Throw exception when the problem report key uses invalid characters."""
        report = problem_report.ProblemReport()
        with (
            io.BytesIO(b"Invalid/key: value") as report_file,
            self.assertRaisesRegex(
                problem_report.InvalidProblemReportKeyError, "key 'Invalid/key'"
            ),
        ):
            report.load(report_file)

    def test_load_invalid_utf8(self) -> None:
        """Throw exception when binary file is invalid UTF-8."""
        report = problem_report.ProblemReport()
        with (
            io.BytesIO(b"\x7fELF\x02\x01\xb0j") as report_file,
            self.assertRaisesRegex(
                problem_report.MalformedProblemReport,
                r"Line '\\x7fELF\\x02\\x01\\\\xb0j' does not contain a colon",
            ),
        ):
            report.load(report_file)

    def test_load_incorrect_padding(self) -> None:
        """Throw exception when base64 encoded data has incorrect padding."""
        report = problem_report.ProblemReport()
        content = (
            b"CoreDump: base64\n"
            b" H4sICAAAAAAC/0NvcmVEdW1wAA==\n"
            b" 7Z0LYFPV/cdP0rQ\n"
        )
        with (
            io.BytesIO(content) as report_file,
            self.assertRaisesRegex(
                problem_report.MalformedProblemReport,
                "^Malformed problem report: Incorrect padding."
                " Is this a proper .crash text file\\?$",
            ),
        ):
            report.load(report_file)

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_load_zstd_compressed_data(self) -> None:
        """Test reading zstd-compressed data."""
        report = problem_report.ProblemReport()
        content = (
            b"CoreDump: base64\n"
            b" KLUv/SROHQIAgoQOEqC7ASCSGCybZRrN7//Hsn7dVyActu7bbMcLaav0RC06\n"
            b" m6fCZ/N7aeOeyqxspRVn88bSx8a8opWQAwEA/8OEAhOOwLA=\n"
        )
        with io.BytesIO(content) as report_file:
            report.load(report_file)
        self.assertEqual(
            report["CoreDump"],
            "sample data that is compressed with zstd"
            " and is long enough for zstd to work.\n",
        )

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_reading_zstd_compressed_value(self) -> None:
        """Test reading zstd-compressed CompressedValue."""
        report = problem_report.ProblemReport()
        content = (
            b"CoreDump: base64\n"
            b" KLUv/SRPHQIA8sQPFKClOoBIEkjIbmfbhiz9/1/96mYFATGiz9JHLfTZY9aV\n"
            b" a9QBRdQz0AkUjbdbVqNQorP3UpGC1NIi9X3UgygBAEyQcYs=\n"
        )
        with io.BytesIO(content) as report_file:
            report.load(report_file, binary="compressed")
        self.assertEqual(
            report["CoreDump"].get_value(),
            b"fake core dump data for testing purposes"
            b" which is long enough to be compressed\n",
        )

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_writing_zstd_compressed_value(self) -> None:
        """Test writing zstd-compressed CompressedValue."""
        compressed_value = problem_report.CompressedValue(
            compressed_value=base64.b64decode(
                b"KLUv/SRCvQEAIsQMEMC3AbLcS0WaZ8rvf69jyw3ghpXr6pBr5i7HiCKroc60"
                b"3+tB4rC/4TV1osHyJUXme6IIAD6PTiM="
            )
        )
        output = io.BytesIO()
        compressed_value.write(output)
        self.assertEqual(
            output.getvalue(),
            b"Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam\n",
        )

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_zstd_compressed_value_length(self) -> None:
        """Test getting length of zstd-compressed CompressedValue."""
        compressed_value = problem_report.CompressedValue(
            compressed_value=base64.b64decode(
                b"KLUv/SRB3QEAwoMMEbDrOJRilklIQlLIJC3OvFUMIIwdaHxrV6aSHEv6rFlP"
                b"/ll/IaXfo19actUkdvm8hggBAAlhlgHbiBMF"
            )
        )
        self.assertEqual(len(compressed_value), 65)

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_len_zstd_compressed_value_nosize(self) -> None:
        """Test len() on zstd-compressed CompressedValue without a size header."""
        compressed_value = problem_report.CompressedValue(
            compressed_value=base64.b64decode(b"KLUv/QBYEQAAe30=")
        )
        self.assertEqual(len(compressed_value), 2)

    @unittest.mock.patch("builtins.__import__")
    def test_zstandard_missing(self, import_mock: MagicMock) -> None:
        """Test reading zstd-compressed data when zstandard is missing."""
        with contextlib.suppress(KeyError):
            sys.modules.pop("zstandard")
        import_mock.side_effect = ImportError("mocked import error")

        report = problem_report.ProblemReport()
        content = (
            b"CoreDump: base64\n"
            b" KLUv/SROHQIAgoQOEqC7ASCSGCybZRrN7//Hsn7dVyActu7bbMcLaav0RC06\n"
            b" m6fCZ/N7aeOeyqxspRVn88bSx8a8opWQAwEA/8OEAhOOwLA=\n"
        )
        expected_message = (
            "Failed to import zstandard library: mocked import error."
            " Please install python3-zstandard."
        )
        with (
            self.assertRaisesRegex(RuntimeError, expected_message),
            io.BytesIO(content) as report_file,
        ):
            report.load(report_file)

    def test_write_fileobj(self) -> None:
        """Write a report with a pointer to a file-like object."""
        tempbin = io.BytesIO(BIN_DATA)
        tempasc = io.BytesIO(b"Hello World")

        pr = problem_report.ProblemReport(date="now!")
        pr["BinFile"] = (tempbin,)
        pr["AscFile"] = (tempasc, False)
        out = io.BytesIO()
        pr.write(out)
        out.seek(0)

        pr = problem_report.ProblemReport()
        pr.load(out)
        self.assertEqual(pr["BinFile"], tempbin.getvalue())
        self.assertEqual(pr["AscFile"], tempasc.getvalue().decode())

    def test_write_empty_fileobj(self) -> None:
        """Write a report with a pointer to a file-like object with
        enforcing non-emptyness."""
        tempbin = io.BytesIO(b"")
        tempasc = io.BytesIO(b"")

        pr = problem_report.ProblemReport(date="now!")
        pr["BinFile"] = (tempbin, True, None, True)
        out = io.BytesIO()
        self.assertRaises(OSError, pr.write, out)

        pr = problem_report.ProblemReport(date="now!")
        pr["AscFile"] = (tempasc, False, None, True)
        out = io.BytesIO()
        self.assertRaises(OSError, pr.write, out)

    @patch("problem_report.CompressedValue.iter_compressed.__defaults__", (48,))
    def test_write_compressed_value(self) -> None:
        """Write a report with a multi-line compressed value."""
        report = problem_report.ProblemReport(date="now!")
        report["BinValue"] = problem_report.CompressedValue(
            compressed_value=b"\x1f\x8b\x08\x00\x00\x00\x00\x00\x02\xff\x0b\xc9HUH\xcc"
            b"+.O-R(\xc9W\xc8\xc9LK\xd5Q(\x01\x8a\x95\xe6e\x96\xa5\x16\x15\x03y\x89y)"
            b"\n\xa9@veIFf^:\x00\xa6\xbfr\x950\x00\x00\x00"
        )
        out = io.BytesIO()
        report.write(out)
        self.assertEqual(
            out.getvalue().decode(),
            textwrap.dedent("""\
                ProblemType: Crash
                Date: now!
                BinValue: base64
                 H4sIAAAAAAAC/wvJSFVIzCsuTy1SKMlXyMlMS9VRKAGKleZllqUWFQN5iXkpCqlA
                 dmVJRmZeOgCmv3KVMAAAAA==
                """),
        )

    def test_read_file(self) -> None:
        """Read a report with binary data."""
        bin_report = textwrap.dedent("""\
            ProblemType: Crash
            Date: now!
            File: base64
             H4sICAAAAAAC/0ZpbGUA
             c3RyhEIGBoYoRiYAM5XUCxAAAAA=
            Foo: Bar
            """).encode()

        # test with reading everything
        pr = problem_report.ProblemReport()
        skipped_keys = pr.load(io.BytesIO(bin_report))
        self.assertEqual(pr["File"], BIN_DATA)
        self.assertEqual(skipped_keys, [])

        # test with skipping binary data
        skipped_keys = pr.load(io.BytesIO(bin_report), binary=False)
        self.assertEqual(skipped_keys, ["File"])

        # test with keeping compressed binary data
        skipped_keys = pr.load(io.BytesIO(bin_report), binary="compressed")
        self.assertEqual(pr["Foo"], "Bar")
        self.assertEqual(skipped_keys, [])
        self.assertTrue(isinstance(pr["File"], problem_report.CompressedValue))
        self.assertEqual(len(pr["File"]), len(BIN_DATA))
        self.assertEqual(pr["File"].get_value(), BIN_DATA)
        self.assertEqual(pr["File"].name, "File")

    def test_read_file_legacy(self) -> None:
        """Read a report with binary data in legacy format without gzip
        header."""
        bin_report = textwrap.dedent("""\
            ProblemType: Crash
            Date: now!
            File: base64
             eJw=
             c3RyxIAMcBAFAG55BXk=
            Foo: Bar
            """).encode()

        # test with reading everything
        pr = problem_report.ProblemReport()
        skipped_keys = pr.load(io.BytesIO(bin_report))
        self.assertEqual(pr["File"], b"AB" * 10 + b"\0" * 10 + b"Z")
        self.assertEqual(skipped_keys, [])

        # test with skipping binary data
        skipped_keys = pr.load(io.BytesIO(bin_report), binary=False)
        self.assertEqual(skipped_keys, ["File"])

        # test with keeping CompressedValues
        skipped_keys = pr.load(io.BytesIO(bin_report), binary="compressed")
        self.assertEqual(skipped_keys, [])
        self.assertEqual(len(pr["File"]), 31)
        self.assertEqual(pr["File"].get_value(), b"AB" * 10 + b"\0" * 10 + b"Z")
        self.assertEqual(pr["File"].name, "File")
        out = io.BytesIO()
        pr["File"].write(out)
        out.seek(0)
        self.assertEqual(out.read(), b"AB" * 10 + b"\0" * 10 + b"Z")

    def test_iter(self) -> None:
        """problem_report.ProblemReport iteration."""
        pr = problem_report.ProblemReport()
        pr["foo"] = "bar"

        keys = list(pr.keys())
        keys.sort()
        self.assertEqual(" ".join(keys), "Date ProblemType foo")

        self.assertEqual(len([k for k in pr if k != "foo"]), 2)

    def test_modify(self) -> None:
        """reading, modifying fields, and writing back."""
        report = textwrap.dedent("""\
            ProblemType: Crash
            Date: now!
            Long:
             xxx
             .
             yyy
            Short: Bar
            File: base64
             H4sICAAAAAAC/0ZpbGUA
             c3RyxIAMcBAFAK/2p9MfAAAA
            """).encode()

        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report), binary="compressed")

        self.assertEqual(pr["Long"], "xxx\n.\nyyy")
        self.assertEqual(pr["Short"], "Bar")
        self.assertEqual(
            pr["File"].get_value(), b"ABABABABABABABABABAB\0\0\0\0\0\0\0\0\0\0Z"
        )

        # write back unmodified
        out = io.BytesIO()
        pr.write(out)
        self.assertEqual(
            out.getvalue().decode(),
            textwrap.dedent("""\
                ProblemType: Crash
                Date: now!
                Long:
                 xxx
                 .
                 yyy
                Short: Bar
                File: base64
                 H4sICAAAAAAC/0ZpbGUAc3RyxIAMcBAFAK/2p9MfAAAA
                """),
        )

        pr["Short"] = "aaa\nbbb"
        pr["Long"] = "123"
        out = io.BytesIO()
        pr.write(out)
        self.assertEqual(
            out.getvalue().decode(),
            textwrap.dedent("""\
                ProblemType: Crash
                Date: now!
                Long: 123
                Short:
                 aaa
                 bbb
                File: base64
                 H4sICAAAAAAC/0ZpbGUAc3RyxIAMcBAFAK/2p9MfAAAA
                """),
        )

    def test_sorted_items(self) -> None:
        """Test ProblemReport.sorted_items()."""
        bin_report = textwrap.dedent("""\
            ProblemType: Crash
            Date: now!
            File: base64
             eJw=
             c3RyxIAMcBAFAG55BXk=
            _MarkForUpload: False
            Architecture: amd64
            ExecutablePath: /usr/bin/python3
            Package: python3.12-minimal
            """).encode()

        report = problem_report.ProblemReport()
        report.load(io.BytesIO(bin_report), binary=False)
        self.assertEqual(
            list(report.sorted_items()),
            [
                ("ExecutablePath", "/usr/bin/python3"),
                ("Package", "python3.12-minimal"),
                ("ProblemType", "Crash"),
                ("Architecture", "amd64"),
                ("Date", "now!"),
            ],
        )

    @unittest.skipUnless(zstandard, "zstandard Python module not available")
    def test_write_mime_binary_values(self) -> None:
        """write_mine() for binary values (gzip and zstd compressed)."""
        report = problem_report.ProblemReport(date="now!")
        report["Data.gz"] = GZIP_BIN_DATA
        report["GzipData"] = problem_report.CompressedValue(
            compressed_value=GZIP_BIN_DATA
        )
        report["ZstdData"] = problem_report.CompressedValue(
            compressed_value=ZSTD_BIN_DATA
        )

        output = io.BytesIO()
        report.write_mime(output)
        output.seek(0)

        message = email.message_from_binary_file(output)
        remaining_parts = message.walk()

        # first part is the multipart container
        part = next(remaining_parts)
        self.assertTrue(part.is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        part = next(remaining_parts)
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "text/plain")
        self.assertEqual(part.get_content_charset(), "utf-8")
        self.assertIsNone(part.get_filename())
        self.assertEqual(
            part.get_payload(decode=True), b"ProblemType: Crash\nDate: now!\n"
        )

        # third part should be the Data.gz as attachment
        part = next(remaining_parts)
        self.assertEqual(part.get_filename(), "Data.gz")
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "application/gzip")
        self.assertIsNone(part.get_content_charset())
        self.assertEqual(part.get_payload(decode=True), GZIP_BIN_DATA)

        # fourth part should be the GzipData as attachment
        part = next(remaining_parts)
        self.assertEqual(part.get_filename(), "GzipData.gz")
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "application/gzip")
        self.assertIsNone(part.get_content_charset())
        self.assertEqual(part.get_payload(decode=True), GZIP_BIN_DATA)

        # fifth part should be the ZstdData as attachment
        part = next(remaining_parts)
        self.assertEqual(part.get_filename(), "ZstdData.zst")
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "application/zstd")
        self.assertIsNone(part.get_content_charset())
        self.assertEqual(part.get_payload(decode=True), ZSTD_BIN_DATA)

        with self.assertRaises(StopIteration):
            next(remaining_parts)

    def test_write_mime_invalid_compressed_binary(self) -> None:
        """write_mine() for invalid compressed binary values."""
        report = problem_report.ProblemReport()
        report["InvalidData.gz"] = b"\0X"

        output = io.BytesIO()
        with self.assertRaisesRegex(
            ValueError, "^Unknown compression for InvalidData.gz$"
        ):
            report.write_mime(output)

    def test_write_mime_json(self) -> None:
        """write_mine() for JSON values."""
        package_data = json.dumps({"elfType": "coredump", "key": "A" * 1000}, indent=2)
        pro_status = json.dumps(
            {
                "_schema_version": "0.1",
                "version": "37.2ubuntu",
                "execution_status": "inactive",
                "execution_details": "No Ubuntu Pro operations are running",
            },
            indent=2,
        )
        report = problem_report.ProblemReport(date="now!")
        report["CoredumpPackageJson"] = package_data
        report["pro-status.json"] = pro_status

        output = io.BytesIO()
        report.write_mime(output)
        output.seek(0)

        message = email.message_from_binary_file(output)
        remaining_parts = message.walk()

        # first part is the multipart container
        part = next(remaining_parts)
        self.assertTrue(part.is_multipart())

        # next part should be an inline text/plain attachments with all short
        # fields
        part = next(remaining_parts)
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "text/plain")
        self.assertEqual(part.get_content_charset(), "utf-8")
        self.assertIsNone(part.get_filename())
        self.assertEqual(
            part.get_payload(decode=True), b"ProblemType: Crash\nDate: now!\n"
        )

        # next part should be the CoredumpPackageJson as attachment
        part = next(remaining_parts)
        self.assertEqual(part.get_filename(), "CoredumpPackage.json")
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "application/json")
        self.assertIsNone(part.get_content_charset())
        self.assertEqual(part.get_payload(decode=True), package_data.encode())

        # next part should be the pro-status.json as attachment
        part = next(remaining_parts)
        self.assertEqual(part.get_filename(), "pro-status.json")
        self.assertFalse(part.is_multipart())
        self.assertEqual(part.get_content_type(), "application/json")
        self.assertIsNone(part.get_content_charset())
        self.assertEqual(part.get_payload(decode=True), pro_status.encode())

        with self.assertRaises(StopIteration):
            next(remaining_parts)

    def test_write_mime_text(self) -> None:
        """write_mime() for text values."""
        pr = problem_report.ProblemReport(date="now!")
        pr["Simple"] = "bar"
        pr["SimpleUTF8"] = "1äö2Φ3".encode()
        pr["SimpleUnicode"] = "1äö2Φ3"
        pr["TwoLineUnicode"] = "pi-π\nnu-η\n"
        pr["TwoLineUTF8"] = "pi-π\nnu-η\n".encode()
        pr["SimpleLineEnd"] = "bar\n"
        pr["TwoLine"] = "first\nsecond\n"
        pr["InlineMargin"] = "first\nsecond\nthird\nfourth\nfifth\n"
        pr["Multiline.txt"] = " foo   bar\nbaz\n  blip  \nline4\nline♥5!!\nłıµ€ ⅝\n"

        # still small enough for inline text
        pr["Largeline"] = "A" * 999
        pr["LargeMultiline"] = f"{'A' * 120}\n{'B' * 90}"

        # too big for inline text, these become attachments
        pr["Hugeline"] = "A" * 10000
        pr["HugeMultiline"] = f"{'A' * 900}\n{'B' * 900}\n{'C' * 900}"
        out = io.BytesIO()
        pr.write_mime(out)
        out.seek(0)

        msg = email.message_from_binary_file(out)
        parts = list(msg.walk())
        self.assertEqual(len(parts), 5)

        # first part is the multipart container
        self.assertTrue(parts[0].is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        self.assertTrue(not parts[1].is_multipart())
        self.assertEqual(parts[1].get_content_type(), "text/plain")
        self.assertEqual(parts[1].get_content_charset(), "utf-8")
        self.assertIsNone(parts[1].get_filename())
        expected = textwrap.dedent(f"""\
            ProblemType: Crash
            Date: now!
            InlineMargin:
             first
             second
             third
             fourth
             fifth
            LargeMultiline:
             {'A' * 120}
             {'B' * 90}
            Largeline: {'A' * 999}
            Simple: bar
            SimpleLineEnd: bar
            SimpleUTF8: 1äö2Φ3
            SimpleUnicode: 1äö2Φ3
            TwoLine:
             first
             second
            TwoLineUTF8:
             pi-π
             nu-η
            TwoLineUnicode:
             pi-π
             nu-η
            """).encode("UTF-8")
        self.assertEqual(parts[1].get_payload(decode=True), expected)

        # third part should be the HugeMultiline: field as attachment
        self.assertTrue(not parts[2].is_multipart())
        self.assertEqual(parts[2].get_content_type(), "text/plain")
        self.assertEqual(parts[2].get_content_charset(), "utf-8")
        self.assertEqual(parts[2].get_filename(), "HugeMultiline.txt")
        self.assertEqual(
            parts[2].get_payload(decode=True), pr["HugeMultiline"].encode()
        )

        # fourth part should be the Hugeline: field as attachment
        self.assertTrue(not parts[3].is_multipart())
        self.assertEqual(parts[3].get_content_type(), "text/plain")
        self.assertEqual(parts[3].get_content_charset(), "utf-8")
        self.assertEqual(parts[3].get_filename(), "Hugeline.txt")
        self.assertEqual(parts[3].get_payload(decode=True), pr["Hugeline"].encode())

        # fifth part should be the Multiline: field as attachment
        self.assertTrue(not parts[4].is_multipart())
        self.assertEqual(parts[4].get_content_type(), "text/plain")
        self.assertEqual(parts[4].get_content_charset(), "utf-8")
        self.assertEqual(parts[4].get_filename(), "Multiline.txt")
        expected = textwrap.dedent(f"""\
             foo   bar
            baz
              blip{'  '}
            line4
            line♥5!!
            łıµ€ ⅝
            """).encode("UTF-8")
        self.assertEqual(parts[4].get_payload(decode=True), expected)

    def test_write_mime_extra_headers(self) -> None:
        """write_mime() with extra headers."""
        pr = problem_report.ProblemReport(date="now!")
        pr["Simple"] = "bar"
        pr["TwoLine"] = "first\nsecond\n"
        out = io.BytesIO()
        pr.write_mime(out, extra_headers={"Greeting": "hello world", "Foo": "Bar"})
        out.seek(0)

        msg = email.message_from_binary_file(out)
        self.assertEqual(msg["Greeting"], "hello world")
        self.assertEqual(msg["Foo"], "Bar")
        parts = list(msg.walk())
        self.assertEqual(len(parts), 2)

        # first part is the multipart container
        self.assertTrue(parts[0].is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        self.assertTrue(not parts[1].is_multipart())
        self.assertEqual(parts[1].get_content_type(), "text/plain")
        self.assertIn(b"Simple: bar", parts[1].get_payload(decode=True))

    def test_write_mime_order(self) -> None:
        """write_mime() with keys ordered."""
        pr = problem_report.ProblemReport(date="now!")
        pr["SecondText"] = "What"
        pr["FirstText"] = "Who"
        pr["FourthText"] = "Today"
        pr["ThirdText"] = "I Don't Know"
        out = io.BytesIO()
        pr.write_mime(
            out,
            priority_fields=[
                "FirstText",
                "SecondText",
                "ThirdText",
                "Unknown",
                "FourthText",
            ],
        )
        out.seek(0)

        msg = email.message_from_binary_file(out)
        parts = list(msg.walk())
        self.assertEqual(len(parts), 2)

        # first part is the multipart container
        self.assertTrue(parts[0].is_multipart())

        # second part should be an inline text/plain attachments with all short
        # fields
        self.assertTrue(not parts[1].is_multipart())
        self.assertEqual(parts[1].get_content_type(), "text/plain")
        self.assertEqual(parts[1].get_content_charset(), "utf-8")
        self.assertIsNone(parts[1].get_filename())
        self.assertEqual(
            parts[1].get_payload(decode=True),
            textwrap.dedent("""\
                FirstText: Who
                SecondText: What
                ThirdText: I Don't Know
                FourthText: Today
                ProblemType: Crash
                Date: now!
                """).encode(),
        )

    def test_updating(self) -> None:
        """new_keys() and write() with only_new=True."""
        pr = problem_report.ProblemReport()
        self.assertEqual(pr.new_keys(), {"ProblemType", "Date"})
        pr.load(io.BytesIO(textwrap.dedent("""\
            ProblemType: Crash
            Date: now!
            Foo: bar
            Baz: blob
            """).encode()))

        self.assertEqual(pr.new_keys(), set())

        pr["Foo"] = "changed"
        pr["NewKey"] = "new new"
        self.assertEqual(pr.new_keys(), {"NewKey"})

        out = io.BytesIO()
        pr.write(out, only_new=True)
        self.assertEqual(out.getvalue(), b"NewKey: new new\n")

    def test_import_dict(self) -> None:
        """Import a dictionary with update()."""
        pr = problem_report.ProblemReport()
        pr["oldtext"] = "Hello world"
        pr["oldbin"] = BIN_DATA
        pr["overwrite"] = "I am crap"

        d = {
            "newtext": "Goodbye world",
            "newbin": "11\000\001\002\xffZZ",
            "overwrite": "I am good",
        }

        pr.update(d)
        self.assertEqual(pr["oldtext"], "Hello world")
        self.assertEqual(pr["oldbin"], BIN_DATA)
        self.assertEqual(pr["newtext"], "Goodbye world")
        self.assertEqual(pr["newbin"], "11\000\001\002\xffZZ")
        self.assertEqual(pr["overwrite"], "I am good")

    def test_load_key_filter(self) -> None:
        """Load a report with filtering keys."""
        report = textwrap.dedent("""\
            ProblemType: Crash
            DataNo: nonono
            GoodFile: base64
             H4sICAAAAAAC/0FmaWxlAA==
             c3RyhEIGBoYoRiYAM5XUCxAAAAA=
            DataYes: yesyes
            BadFile: base64
             H4sICAAAAAAC/0ZpbGUA
             S8vPZ0hKLAIACq50HgcAAAA=
            """).encode()
        pr = problem_report.ProblemReport()
        pr.load(io.BytesIO(report), key_filter=["DataYes", "GoodFile"])
        self.assertEqual(pr["DataYes"], "yesyes")
        self.assertEqual(pr["GoodFile"], BIN_DATA)
        self.assertEqual(sorted(pr.keys()), ["DataYes", "GoodFile"])

    def test_get_on_disk_size(self) -> None:
        """Test CompressedValue.get_on_disk_size()."""
        compressed_value = problem_report.CompressedValue(b"somedata")
        self.assertEqual(compressed_value.get_compressed_size(), 28)
        assert compressed_value.compressed_value is not None
        base64_encoded = base64.b64encode(compressed_value.compressed_value)
        self.assertEqual(compressed_value.get_on_disk_size(), len(base64_encoded))


class TestEntryParser(unittest.TestCase):
    # pylint: disable=protected-access
    """Test _EntryParser class."""

    def test_parse(self) -> None:
        """Test parsing a report file with _EntryParser."""
        report_file = io.BytesIO(textwrap.dedent("""\
            First: single line
            Second: multi
             line
            """).encode())
        entries = []
        for entry in problem_report._EntryParser(report_file):
            entries.append(list(entry))
        self.assertEqual(
            entries, [[b"First: single line\n"], [b"Second: multi\n", b" line\n"]]
        )

    def test_skip_entries(self) -> None:
        """Test skipping reading one entry."""
        report_file = io.BytesIO(textwrap.dedent("""\
            First: this entry
             will be skipped
            Second: single line
            """).encode())
        iterator = problem_report._EntryParser(report_file)
        next(iterator)
        self.assertEqual(list(next(iterator)), [b"Second: single line\n"])

    def test_skip_partial_entries(self) -> None:
        """Test skipping reading one entry."""
        report_file = io.BytesIO(textwrap.dedent("""\
            First: this line is read,
             but these lines
             will be skipped
            Second: single line
            """).encode())
        iterator = problem_report._EntryParser(report_file)
        self.assertEqual(next(next(iterator)), b"First: this line is read,\n")
        self.assertEqual(list(next(iterator)), [b"Second: single line\n"])

    def test_skip_last_entry(self) -> None:
        """Test skipping reading the last entry."""
        report_file = io.BytesIO(textwrap.dedent("""\
            First: this line is read
            Second: this line is skipped
            """).encode())
        iterator = problem_report._EntryParser(report_file)
        self.assertEqual(list(next(iterator)), [b"First: this line is read\n"])
        next(iterator)
        with self.assertRaises(StopIteration):
            next(iterator)



# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for apport.packaging_impl.apt_dpkg."""

import pathlib
import pickle
import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

import apt
import apt.package

from apport.packaging_impl.apt_dpkg import (
    _map_mirror_to_arch,
    _parse_deb822_sources,
    _read_mirror_file,
    impl,
)


@unittest.mock.patch(
    "apport.packaging_impl.apt_dpkg._AptDpkgPackageInfo.get_os_version",
    MagicMock(return_value=("Ubuntu", "22.04")),
)
class TestPackagingAptDpkg(unittest.TestCase):
    """Unit tests for apport.packaging_impl.apt_dpkg."""

    maxDiff = None

    def setUp(self) -> None:
        # Clear APT cache to allow mocking it
        # pylint: disable-next=protected-access
        impl._apt_cache = None

    def tearDown(self) -> None:
        # Clear APT cache to clear cached mocks
        # pylint: disable-next=protected-access
        impl._apt_cache = None

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_no_candidate(self, apt_cache_mock: MagicMock) -> None:
        """is_distro_package() for package that has no candidate."""
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, installed=None, candidate=None
        )
        self.assertEqual(impl.is_distro_package("adduser"), False)
        getitem_mock.assert_called_once_with("adduser")

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_no_installed_version(
        self, apt_cache_mock: MagicMock
    ) -> None:
        """is_distro_package() for not installed package."""
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package,
            installed=MagicMock(spec=apt.package.Version, version=None),
        )
        self.assertEqual(impl.is_distro_package("7zip"), False)
        getitem_mock.assert_called_once_with("7zip")

    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_ppa(self, apt_cache_mock: MagicMock) -> None:
        """is_distro_package() for a PPA package."""
        version = MagicMock(
            spec=apt.package.Version,
            origins=[MagicMock(spec=apt.package.Origin, origin="LP-PPA")],
            version="0.5.0-0ppa1",
        )
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, candidate=version, installed=version
        )
        self.assertEqual(impl.is_distro_package("bdebstrap"), False)
        getitem_mock.assert_called_once_with("bdebstrap")

    @unittest.mock.patch("os.path.exists")
    @unittest.mock.patch("apt.Cache", spec=apt.Cache)
    def test_is_distro_package_system_image(
        self, apt_cache_mock: MagicMock, exists_mock: MagicMock
    ) -> None:
        """is_distro_package() for a system image without cache."""
        version = MagicMock(
            spec=apt.package.Version,
            origins=[MagicMock(spec=apt.package.Origin, origin="")],
            version="2.23.1-0ubuntu4",
        )
        getitem_mock = apt_cache_mock.return_value.__getitem__
        getitem_mock.return_value = MagicMock(
            spec=apt.Package, candidate=version, installed=version
        )
        exists_mock.return_value = True
        self.assertEqual(impl.is_distro_package("apport"), True)
        getitem_mock.assert_called_once_with("apport")
        exists_mock.assert_called_once_with("/etc/system-image/channel.ini")

    def test_get_library_paths_unknown_arch(self) -> None:
        """get_library_paths() for unknown architecture."""
        with self.assertRaisesRegex(NotImplementedError, "architecture osprey64"):
            impl.get_library_paths("osprey64")

    def test_map_mirror_to_arch_ports_to_primary(self) -> None:
        """Test _map_mirror_to_arch() to map ports to primary."""
        self.assertEqual(
            _map_mirror_to_arch("http://ports.ubuntu.com/ubuntu-ports/", "amd64"),
            "http://archive.ubuntu.com/ubuntu",
        )

    def test_map_mirror_to_arch_country_ports_to_primary(self) -> None:
        """Test _map_mirror_to_arch() to map country ports to primary."""
        self.assertEqual(
            _map_mirror_to_arch("http://de.ports.ubuntu.com/ubuntu-ports", "amd64"),
            "http://de.archive.ubuntu.com/ubuntu",
        )

    def test_map_mirror_to_arch_ports_unchanged(self) -> None:
        """Test _map_mirror_to_arch() to keep ports unchanged."""
        uri = "http://de.ports.ubuntu.com/ubuntu-ports"
        self.assertEqual(_map_mirror_to_arch(uri, "ppc64el"), uri)

    def test_map_mirror_to_arch_primary_to_ports(self) -> None:
        """Test _map_mirror_to_arch() to map primary to ports."""
        self.assertEqual(
            _map_mirror_to_arch("http://de.archive.ubuntu.com/ubuntu/", "s390x"),
            "http://de.ports.ubuntu.com/ubuntu-ports",
        )

    def test_map_mirror_to_arch_primary_unchanged(self) -> None:
        """Test _map_mirror_to_arch() to keep ports unchanged."""
        uri = "http://de.archive.ubuntu.com/ubuntu"
        self.assertEqual(_map_mirror_to_arch(uri, "amd64"), uri)

    @unittest.mock.patch(
        "builtins.open",
        new_callable=unittest.mock.mock_open,
        read_data="""
# Some documentation in the beginning

Types: deb deb-src
URIs: http://example.com
Suites: foo foo-bar
Components: main


Types: deb
URIs: http://example2.com
Suites:
 baz
Components: main


""",
    )
    def test_parse_deb822_sources_extra_lines(self, mock_file: MagicMock) -> None:
        """Test _parse_deb822_sources with multiple lines separating blocks."""
        entries = _parse_deb822_sources("foo_bar.sources")
        mock_file.assert_called_with("foo_bar.sources", encoding="utf-8")
        self.assertEqual(len(entries), 2)
        self.assertEqual(entries[0].uris[0], "http://example.com")
        self.assertEqual(entries[1].suites[0], "baz")

    def test_read_mirror_file(self) -> None:
        """Test _read_mirror_file with config from GitHub CI."""
        with tempfile.NamedTemporaryFile("w") as mirror_file:
            mirror_file.write(
                "http://azure.archive.ubuntu.com/ubuntu/\tpriority:1\n"
                "http://archive.ubuntu.com/ubuntu/\tpriority:2\n"
                "http://security.ubuntu.com/ubuntu/\tpriority:3\n"
            )
            mirror_file.flush()
            mirrors = _read_mirror_file(f"mirror+file:{mirror_file.name}")
        self.assertEqual(
            mirrors,
            [
                "http://azure.archive.ubuntu.com/ubuntu/",
                "http://archive.ubuntu.com/ubuntu/",
                "http://security.ubuntu.com/ubuntu/",
            ],
        )

    @unittest.mock.patch.object(impl, "_get_file2pkg_mapping")
    @unittest.mock.patch.object(impl, "_save_contents_mapping", MagicMock())
    def test_get_file_package_uninstalled_usrmerge(
        self, _get_file2pkg_mapping_mock: MagicMock
    ) -> None:
        """get_file_package() on uninstalled usrmerge packages."""
        # Data from Ubuntu 24.04 (noble)
        _get_file2pkg_mapping_mock.return_value = {
            "usr/lib/x86_64-linux-gnu/ld-linux-x86-64.so.2": "libc6",
            "usr/lib/x86_64-linux-gnu/libc.so.6": "libc6",
            "usr/libx32/libc.so.6": "libc6-x32",
        }

        pkg = impl.get_file_package(
            "/lib/x86_64-linux-gnu/libc.so.6", True, "/map_cachedir", arch="amd64"
        )

        self.assertEqual(pkg, "libc6")
        _get_file2pkg_mapping_mock.assert_called_with(
            "/map_cachedir", impl.get_distro_codename(), "amd64"
        )

    def test_contents_mapping_load_legacy_resolute(self) -> None:
        """Test _contents_mapping() discarding cache file from Apport 2.34.0."""
        with tempfile.TemporaryDirectory() as workdir:
            legacy_file2pkg = {
                b"release": b"resolute",
                b"arch": b"amd64",
                b"path/to/file": b"package1",
            }
            mapping_path = (
                pathlib.Path(workdir) / "contents_mapping-resolute-amd64.pickle"
            )
            with mapping_path.open("wb") as mapping_file:
                pickle.dump(legacy_file2pkg, mapping_file)

            # pylint: disable-next=protected-access
            file2pkg = impl._contents_mapping(workdir, "resolute", "amd64")
            self.assertEqual(file2pkg, {"release": "resolute", "arch": "amd64"})

    def test_contents_skip_xenial_header(self) -> None:
        """Test _update_given_file2pkg_mapping skipping xenial Contents header."""
        # Header taken from
        # http://archive.ubuntu.com/ubuntu/dists/xenial/Contents-amd64.gz
        contents = """\
This file maps each file available in the Ubuntu
system to the package from which it originates.  It includes packages
from the DIST distribution for the ARCH architecture.

You can use this list to determine which package contains a specific
file, or whether or not a specific file is available.  The list is
updated weekly, each architecture on a different day.

When a file is contained in more than one package, all packages are
listed.  When a directory is contained in more than one package, only
the first is listed.

The best way to search quickly for a file is with the Unix `grep'
utility, as in `grep <regular expression> CONTENTS':

 $ grep nose Contents
 etc/nosendfile                                          net/sendfile
 usr/X11R6/bin/noseguy                                   x11/xscreensaver
 usr/X11R6/man/man1/noseguy.1x.gz                        x11/xscreensaver
 usr/doc/examples/ucbmpeg/mpeg_encode/nosearch.param     graphics/ucbmpeg
 usr/lib/cfengine/bin/noseyparker                        admin/cfengine

This list contains files in all packages, even though not all of the
packages are installed on an actual system at once.  If you want to
find out which packages on an installed Debian system provide a
particular file, you can use `dpkg --search <filename>':

 $ dpkg --search /usr/bin/dselect
 dpkg: /usr/bin/dselect


FILE                                                    LOCATION
:sexsend:sexget:					    universe/web/fex
bin/afio						    multiverse/utils/afio
bin/archdetect						    utils/archdetect-deb
"""
        file2pkg: dict[str, str] = {}
        open_mock = unittest.mock.mock_open(read_data=contents)
        with unittest.mock.patch("gzip.open", open_mock):
            # pylint: disable-next=protected-access
            impl._update_given_file2pkg_mapping(file2pkg, "/fake_Contents", "xenial")

        self.assertEqual(
            file2pkg,
            {
                ":sexsend:sexget:": "fex",
                "bin/afio": "afio",
                "bin/archdetect": "archdetect-deb",
            },
        )
        open_mock.assert_called_once_with("/fake_Contents", "rt")

    def test_contents_path_filering(self) -> None:
        """Test _update_given_file2pkg_mapping to ignore unrelevant files."""
        # Test content taken from
        # http://archive.ubuntu.com/ubuntu/dists/noble/Contents-amd64.gz
        contents = """\
bin/ip							    net/iproute2
boot/ipxe.efi						    admin/grub-ipxe
etc/dput.cf						    devel/dput
lib/modules/6.8.0-31-generic/kernel/crypto/lz4.ko.zst	\
    kernel/linux-modules-6.8.0-31-generic
lib/nut/clone						    admin/nut-server
sbin/hdparm						    admin/hdparm
usr/Brother/inf/braddprinter				    multiverse/text/brother-lpr-drivers-laser
usr/aarch64-linux-gnu/include/ar.h			    libdevel/libc6-dev-arm64-cross
usr/bin/git						    vcs/git
usr/bin/xz						    utils/xz-utils
usr/games/etr						    universe/games/extremetuxracer
usr/include/apache2/os.h				    httpd/apache2-dev
usr/include/x86_64-linux-gnu/bits/endian.h		    libdevel/libc6-dev
usr/lib/7zip/7z.so					    universe/utils/7zip
usr/lib/debug/.build-id/31/1c9c9b30d6991fb903ab459173c66eb8e7e895.debug debug/libc6-dbg
usr/libexec/coreutils/libstdbuf.so			    utils/coreutils
usr/libx32/ld.so					    libs/libc6-x32
usr/sbin/zic						    libs/libc-bin
usr/share/cargo/registry/libc-0.2.152/src/lib.rs	    universe/rust/librust-libc-dev
usr/share/dicom3tools/gen.so				    universe/graphics/dicom3tools
usr/share/doc/0install					    universe/admin/0install
usr/share/gimp/2.0/help/de/glossary.html		    universe/doc/gimp-help-de
usr/share/gocode/src/launchpad.net/mgo			    universe/devel/golang-gopkg-mgo.v2-dev
usr/share/help/C/eog/default.page			    gnome/eog
usr/share/icons/gnome-colors-common/32x32/apps/konsole.png\
  universe/gnome/gnome-colors-common
usr/share/libreoffice/help/en-US/noscript.html		    doc/libreoffice-help-en-us
usr/share/locale/de/LC_MESSAGES/apt.mo			    admin/apt
usr/share/man/de/man1/man.1.gz				    doc/man-db
usr/share/texlive/index.html				    universe/tex/texlive-base
usr/src/broadcom-sta.tar.xz				    restricted/admin/broadcom-sta-source
var/lib/ieee-data/iab.txt				    net/ieee-data
"""

        file2pkg: dict[str, str] = {}
        open_mock = unittest.mock.mock_open(read_data=contents)
        with unittest.mock.patch("gzip.open", open_mock):
            # pylint: disable-next=protected-access
            impl._update_given_file2pkg_mapping(file2pkg, "Contents-amd64", "noble")

        self.assertEqual(
            file2pkg,
            {
                "bin/ip": "iproute2",
                "boot/ipxe.efi": "grub-ipxe",
                "etc/dput.cf": "dput",
                "lib/nut/clone": "nut-server",
                "sbin/hdparm": "hdparm",
                "usr/Brother/inf/braddprinter": "brother-lpr-drivers-laser",
                "usr/bin/git": "git",
                "usr/lib/debug/.build-id/31/1c9c9b30d6991fb903ab459173c66eb8e7e895"
                ".debug": "libc6-dbg",
                "usr/bin/xz": "xz-utils",
                "usr/games/etr": "extremetuxracer",
                "usr/lib/7zip/7z.so": "7zip",
                "usr/libexec/coreutils/libstdbuf.so": "coreutils",
                "usr/libx32/ld.so": "libc6-x32",
                "usr/sbin/zic": "libc-bin",
                "usr/share/dicom3tools/gen.so": "dicom3tools",
                "var/lib/ieee-data/iab.txt": "ieee-data",
            },
        )
        open_mock.assert_called_once_with("Contents-amd64", "rt")

    def test_contents_parse_path_with_spaces(self) -> None:
        """Test _update_given_file2pkg_mapping to parse Contents file correctly."""
        # Test content taken from
        # http://archive.ubuntu.com/ubuntu/dists/noble/Contents-amd64.gz
        contents = (
            "usr/lib/iannix/Tools/JavaScript Library.js\t\t    "
            "universe/sound/iannix\n"
            "usr/lib/python3/dist-packages/ilorest/extensions/BIOS COMMANDS"
            "/__init__.py universe/python/ilorest\n"
        )

        file2pkg: dict[str, str] = {}
        open_mock = unittest.mock.mock_open(read_data=contents)
        with unittest.mock.patch("gzip.open", open_mock):
            # pylint: disable-next=protected-access
            impl._update_given_file2pkg_mapping(file2pkg, "Contents-amd64", "noble")

        self.assertEqual(
            file2pkg,
            {
                "usr/lib/iannix/Tools/JavaScript Library.js": "iannix",
                "usr/lib/python3/dist-packages/ilorest/extensions/BIOS COMMANDS"
                "/__init__.py": "ilorest",
            },
        )
        open_mock.assert_called_once_with("Contents-amd64", "rt")


class TestPackaging(unittest.TestCase):
    """Unit tests for apport.packaging.PackageInfo class."""

    def setUp(self) -> None:
        # pylint: disable-next=protected-access
        impl._os_version = None

    def tearDown(self) -> None:
        # pylint: disable-next=protected-access
        impl._os_version = None

    @unittest.mock.patch("platform.freedesktop_os_release")
    def test_get_os_version_debian_testing(self, os_release_mock: MagicMock) -> None:
        """Test get_os_version() for Debian testing (no VERSION_ID set)."""
        # platform.freedesktop_os_release() data from base-files 13.7
        os_release_mock.return_value = {
            "NAME": "Debian GNU/Linux",
            "ID": "debian",
            "PRETTY_NAME": "Debian GNU/Linux trixie/sid",
            "VERSION_CODENAME": "trixie",
            "HOME_URL": "https://www.debian.org/",
            "SUPPORT_URL": "https://www.debian.org/support",
            "BUG_REPORT_URL": "https://bugs.debian.org/",
        }
        self.assertEqual(impl.get_os_version(), ("Debian", "n/a"))
        os_release_mock.assert_called_once_with()



# Copyright (C) 2024 Canonical Ltd.
# Author: Simon Chopin <simon.chopin@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for apport.crash_impl.launchpad"""

import io
import re
import unittest.mock

from apport.crashdb_impl.launchpad import CrashDatabase, upload_blob
from apport.report import Report


def python_crash() -> Report:
    """Generate a report that looks like a Python crash"""
    report = Report("Crash")
    report["Package"] = "python-goo 3epsilon1"
    report["SourcePackage"] = "pygoo"
    report["PackageArchitecture"] = "all"
    report["DistroRelease"] = "Ubuntu 24.04"
    report["ExecutablePath"] = "/usr/bin/pygoo"
    report["Traceback"] = """Traceback (most recent call last):
  File "test.py", line 7, in <module>
    print(_f(5))
  File "test.py", line 5, in _f
    return g_foo00(x+1)
  File "test.py", line 2, in g_foo00
    return x/0
ZeroDivisionError: integer division or modulo by zero"""
    return report


def native_crash() -> Report:
    """Generate a report that looks like a native binary crash"""
    report = Report("Crash")
    report["Signal"] = "6"
    report["SignalName"] = "SIGABRT"
    report["Package"] = "bash"
    report["SourcePackage"] = "bash"
    report["DistroRelease"] = "Ubuntu 24.04"
    report["PackageArchitecture"] = "i386"
    report["Architecture"] = "amd64"
    report["ExecutablePath"] = "/bin/bash"
    report["CoreDump"] = "/var/lib/apport/coredump/core.bash"
    report["AssertionMessage"] = "foo.c:42 main: i > 0"
    return report


def test_python_crash_headers() -> None:
    # pylint: disable=protected-access
    """Test _generate_upload_headers in case of a Python crash"""
    crashdb = CrashDatabase(None, {"distro": "ubuntu"})
    report = python_crash()
    headers = crashdb._generate_upload_headers(report)

    assert "need-duplicate-check" in headers["Tags"].split(" ")
    assert headers.get("Private") == "yes"


def test_native_crash_headers() -> None:
    # pylint: disable=protected-access
    """Test _generate_upload_headers in case of a native crash"""
    crashdb = CrashDatabase(None, {"distro": "ubuntu"})
    report = native_crash()
    headers = crashdb._generate_upload_headers(report)

    assert "i386" in headers["Tags"].split(" ")
    assert "need-i386-retrace" in headers["Tags"].split(" ")
    assert headers.get("Private") == "yes"


def test_private_bug_headers() -> None:
    # pylint: disable=protected-access
    """Test _generate_upload_headers for a bug in a package for which
    the hook explicitly says it should be private"""
    crashdb = CrashDatabase(None, {"distro": "ubuntu"})
    report = Report("Bug")
    report["Package"] = "apport"
    report["SourcePackage"] = "apport"
    report["PackageArchitecture"] = "all"
    report["Architecture"] = "amd64"
    report["DistroRelease"] = "Ubuntu 24.04"
    report["LaunchpadPrivate"] = "yes"
    headers = crashdb._generate_upload_headers(report)

    assert headers.get("Private") == "yes"
    assert not re.search(r"need-[a-z0-9]+-retrace", headers.get("Tags", ""))


@unittest.mock.patch("urllib.request.build_opener")
def test_upload_blob_conform_to_lp(builder_mock: unittest.mock.MagicMock) -> None:
    """Test that upload_blob does NOT encodes form data with CRLF as line separators,
    despite HTTP 1.1 spec"""
    builder_mock.return_value.open.return_value.info.return_value = {
        "X-Launchpad-Blob-Token": 42
    }

    data = io.BytesIO(b"foobarbaz")
    result = upload_blob(data, hostname="invalid.host")

    builder_mock.assert_called_once()
    builder_mock.return_value.open.assert_called_once()
    req = builder_mock.return_value.open.call_args[0][0]
    assert b"foobarbaz" in req.data
    # Check that we have LF-separated headers (because LP expects LF
    # line separators, see LP: #2097632)
    assert re.match(rb"^([-A-Za-z]+: [^\r\n]+\n)+\n", req.data)
    assert result == 42



"""Integration tests for the Qt Apport user interface."""

# Copyright (C) 2025 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
# SPDX-License-Identifier: GPL-2.0-or-later

import io
from unittest.mock import patch

import pytest

from tests.helper import import_module_from_file
from tests.paths import get_data_directory


def test_no_environment_variables() -> None:
    """Test launching apport-kde without any environment variables set."""
    apport_kde_path = get_data_directory("kde") / "apport-kde"
    with patch("sys.stderr", new_callable=io.StringIO) as stderr:
        try:
            apport_kde = import_module_from_file(apport_kde_path)
        except SystemExit:
            pytest.skip(stderr.getvalue().strip())
    with (
        patch.dict("os.environ", {}, clear=True),
        patch("sys.stderr", new_callable=io.StringIO) as stderr,
        pytest.raises(SystemExit) as error,
    ):
        apport_kde.main([str(apport_kde_path)])

    assert (
        stderr.getvalue() == "ERROR: This program needs a running display server"
        ' session. Please see "man apport-cli" for a command line version of Apport.\n'
    )
    assert error.value.code == 1



# You do not want to know why this empty file exist! You have been warned!
#
# This empty module exist to be imported by `apport/__init__.py` before
# `apport/__init__.py` defines the `packaging` object instance. This makes
# following code work:
#
# >>> import apport.packaging
# >>> type(apport.packaging)
# apport.packaging_impl.apt_dpkg._AptDpkgPackageInfo
#
# Importing `apport.packaging` will import `apport` first. This will import
# `apport.packaging` and shadow it.



# Copyright (C) 2026 Canonical Ltd.
# Author: Benjamin Drung <bdrung@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for the apport.procutils module."""

from unittest.mock import mock_open, patch

import pytest

from apport.procutils import Meminfo, parse_meminfo


def test_parse_meminfo() -> None:
    """Test parse_meminfo() reading all values successfully."""
    open_mock = mock_open(
        read_data=(
            "MemTotal:       64249228 kB\n"
            "MemFree:        25066812 kB\n"
            "MemAvailable:   40437820 kB\n"
            "Buffers:          587236 kB\n"
            "Cached:         19694816 kB\n"
            "Writeback:             0 kB\n"
        )
    )
    with patch("builtins.open", open_mock):
        meminfo = parse_meminfo()
    assert meminfo == Meminfo(64249228, 25066812, 40437820, 19694816, 0)


def test_parse_meminfo_missing_key() -> None:
    """Test parse_meminfo() failing to read the requested keys."""
    open_mock = mock_open(
        read_data=(
            "MemTotal:       64249228 kB\n"
            "MemFree:        23434628 kB\n"
            "MemAvailable:   38809072 kB\n"
            "Buffers:          589372 kB\n"
            "Cached:         20099444 kB\n"
        )
    )
    with patch("builtins.open", open_mock), pytest.raises(KeyError) as exc_info:
        parse_meminfo()
    exc_info.match("Writeback not found in /proc/meminfo")



# Copyright (C) 2025 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for functions in the base apport module."""

from apport import Report, packaging


def test_report() -> None:
    """Test using Report imported from the base apport module."""
    report = Report()
    assert report["ProblemType"] == "Crash"


def test_packaging() -> None:
    """Test using packaging imported from the base apport module."""
    assert packaging.get_source("gzip") == "gzip"



"""Crash database implementation for Github."""

# Copyright (C) 2022 - 2022 Canonical Ltd.
# Author: Eduard Gómez Escanell <edu.gomez.escandell@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

import json
import time
import urllib.error
import urllib.parse
import urllib.request
from collections.abc import Callable, Mapping
from dataclasses import dataclass
from typing import Any, Self

import apport.crashdb
import apport.report


class Github:
    """Wrapper around Github API, used to log in and post issues."""

    __last_request: float = time.time()

    def __init__(
        self, client_id: str, message_callback: Callable | None = None
    ) -> None:
        self.__client_id = client_id
        self.__authentication_data: dict[str, str] | None = None
        self.__access_token = None
        self.__cooldown = 0.0
        self.__expiry = 0.0
        self.message_callback = message_callback

    def _post(self, url: str, data: str) -> Any:
        """Posts the given data to the given URL.
        Uses auth token if available"""
        headers = {"Accept": "application/vnd.github.v3+json"}
        if self.__access_token:
            headers["Authorization"] = f"token {self.__access_token}"
        request = urllib.request.Request(
            url, data=data.encode("utf-8"), headers=headers, method="POST"
        )
        try:
            with urllib.request.urlopen(request, timeout=30.0) as response:
                return json.loads(response.read())
        except urllib.error.URLError:
            if self.message_callback:
                self.message_callback(
                    "Failed connection",
                    f"Failed connection to {url}.\n"
                    f"Please check your internet connection and try again.",
                )
            raise
        finally:
            self.__last_request = time.time()

    def api_authentication(self, url: str, data: dict) -> Any:
        """Authenticate against the GitHub API."""
        return self._post(url, urllib.parse.urlencode(data))

    def api_open_issue(self, owner: str, repo: str, data: dict) -> Any:
        """Open a new issue on the GitHub project."""
        url = f"https://api.github.com/repos/{owner}/{repo}/issues"
        return self._post(url, json.dumps(data))

    def __enter__(self) -> Self:
        """Enters login process. At exit, login process ends."""
        data = {"client_id": self.__client_id, "scope": "public_repo"}
        url = "https://github.com/login/device/code"
        response = self.api_authentication(url, data)

        prompt = (
            "Posting an issue requires a Github account. If you have "
            "one, please follow these steps to log in.\n"
            "\n"
            "Open the following URL. When requested, write this code "
            "to enable apport to open an issue.\n"
            "URL:  {url}\n"
            "Code: {code}"
        )

        url = response["verification_uri"]
        code = response["user_code"]

        if self.message_callback:
            self.message_callback("Login required", prompt.format(url=url, code=code))

        self.__authentication_data = {
            "client_id": self.__client_id,
            "device_code": f'{response["device_code"]}',
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        }
        self.__cooldown = response["interval"]
        self.__expiry = int(response["expires_in"]) + time.time()

        return self

    def __exit__(self, *_: object) -> None:
        self.__authentication_data = None
        self.__cooldown = 0.0
        self.__expiry = 0.0

    def authentication_complete(self) -> bool:
        """Asks Github if the user has logged in already.
        It respects the wait-time requested by Github.
        """
        if not self.__authentication_data:
            raise RuntimeError(
                "Authentication not started. Use a with statement to do so"
            )

        current_time = time.time()
        waittime = self.__cooldown - (current_time - self.__last_request)
        if current_time + waittime > self.__expiry:
            if self.message_callback:
                self.message_callback(
                    "Failed login", "Github authentication expired. Please try again."
                )
            raise RuntimeError("Github authentication expired")
        if waittime > 0:
            time.sleep(waittime)  # Avoids spamming the API

        url = "https://github.com/login/oauth/access_token"
        response = self.api_authentication(url, self.__authentication_data)

        if "error" in response:
            if response["error"] == "authorization_pending":
                return False
            if response["error"] == "slow_down":
                self.__cooldown = int(response["interval"])
                return False
            raise RuntimeError(f"Unknown error from Github: {response}")
        if "access_token" in response:
            self.__access_token = response["access_token"]
            return True
        raise RuntimeError(f"Unknown response from Github: {response}")


@dataclass(frozen=True)
class IssueHandle:  # pylint: disable=missing-class-docstring
    url: str


class CrashDatabase(apport.crashdb.CrashDatabase):
    """Github crash database.
    This is a Apport CrashDB implementation for interacting with Github issues
    """

    def __init__(self, auth_file: str | None, options: dict[str, Any]) -> None:
        """Initialize some variables. Login is delayed until necessary."""
        apport.crashdb.CrashDatabase.__init__(self, auth_file, options)
        self.repository_owner = options["repository_owner"]
        self.repository_name = options["repository_name"]
        self.app_id = options["github_app_id"]
        self.labels = set(options["labels"])
        self.issue_url = None
        self.github: Github | None = None

    def _format_report(self, report: Mapping[str, Any]) -> dict[str, str | list[str]]:
        """Formats report info as markdown and creates Github issue JSON."""
        body_markdown = ""
        for key, value in report.items():
            body_markdown += f"**{key}**\n{value}\n\n"

        return {
            "title": "Issue submitted via apport",
            "body": body_markdown,
            "labels": list(self.labels),
        }

    def _github_login(self, user_message_callback: Callable | None = None) -> Github:
        with Github(self.app_id, user_message_callback) as github:
            while not github.authentication_complete():
                pass
            return github

    def upload(
        self,
        report: Mapping[str, Any],
        progress_callback: Callable | None = None,
        user_message_callback: Callable | None = None,
    ) -> IssueHandle:
        """Upload given problem report return a handle for it.
        In Github, we open an issue.
        """
        assert self.accepts(report)

        self.github = self._github_login(user_message_callback)

        if self.github is None:
            raise RuntimeError("Failed to login to Github")

        data = self._format_report(report)
        if not (self.repository_name is None and self.repository_owner is None):
            response = self.github.api_open_issue(
                self.repository_owner, self.repository_name, data
            )
        elif "SnapGitOwner" in report and "SnapGitName" in report:
            response = self.github.api_open_issue(
                report["SnapGitOwner"], report["SnapGitName"], data
            )
        else:
            raise RuntimeError(
                "Couldn't determine which repository to file the report in"
            )

        return IssueHandle(url=response["html_url"])

    def get_comment_url(self, report: apport.report.Report, handle: IssueHandle) -> str:
        """Return a URL that should be opened after report has been uploaded
        and upload() returned handle.
        """
        return handle.url

    def _mark_dup_checked(self, crash_id, report):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def can_update(self, crash_id: int) -> bool:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def close_duplicate(self, report, crash_id, master_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def download(self, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def duplicate_of(self, crash_id: int) -> int | None:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_affected_packages(self, crash_id: int) -> list[str]:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_distro_release(self, crash_id: int) -> str:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_dup_unchecked(self) -> set[int]:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_fixed_version(self, crash_id: int) -> str | None:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_id_url(self, report, crash_id):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_unfixed(self) -> set[int]:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def get_unretraced(self) -> set[int]:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def is_reporter(self, crash_id: int) -> bool:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def mark_regression(self, crash_id: int, master: int) -> None:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def mark_retrace_failed(
        self, crash_id: int, invalid_msg: str | None = None
    ) -> None:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    def mark_retraced(self, crash_id: int) -> None:
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )

    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def update(
        self,
        crash_id,
        report,
        comment,
        change_description=False,
        attachment_comment=None,
        key_filter=None,
    ):
        raise NotImplementedError(
            "This method is not relevant for Github database implementation."
        )



# Copyright (C) 2022 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Command line Apport user interface tests."""

import gzip
import io
import os
import pathlib
import tempfile
import unittest
import unittest.mock
from gettext import gettext as _

import apport.report
from problem_report import CompressedFile
from tests.helper import import_module_from_file, skip_if_command_is_missing
from tests.paths import is_local_source_directory, local_test_environment

if is_local_source_directory():
    APPORT_CLI_PATH = "bin/apport-cli"
else:
    APPORT_CLI_PATH = "/usr/bin/apport-cli"
apport_cli = import_module_from_file(pathlib.Path(APPORT_CLI_PATH))


class TestApportCli(unittest.TestCase):
    # pylint: disable=missing-function-docstring
    """Test apport-cli."""

    orig_environ: dict[str, str]

    @classmethod
    def setUpClass(cls) -> None:
        cls.orig_environ = os.environ.copy()
        os.environ |= local_test_environment()

    @classmethod
    def tearDownClass(cls) -> None:
        os.environ.clear()
        os.environ.update(cls.orig_environ)

    def setUp(self) -> None:
        self.app = apport_cli.CLIUserInterface([APPORT_CLI_PATH])
        self.app.report = apport.report.Report()
        self.app.report.add_os_info()
        self.app.report["ExecutablePath"] = "/bin/bash"
        self.app.report["Signal"] = "11"
        self.app.report["CoreDump"] = b"\x01\x02"
        self.app.report["LongString"] = f"l{'o' * 1_042_000}ng"

    @skip_if_command_is_missing("/usr/bin/sensible-pager")
    def test_ui_update_view(self) -> None:
        with tempfile.NamedTemporaryFile(prefix="apport_") as temp:
            with gzip.open(temp.name, "wb") as f_out:
                f_out.write(b"some uncompressed data")
            self.app.report["CompressedFile"] = CompressedFile(temp.name)

            read_fd, write_fd = os.pipe()
            with os.fdopen(write_fd, "w", buffering=1) as stdout:
                self.app.ui_update_view(stdout=stdout)
            with os.fdopen(read_fd, "r") as pipe:
                report = pipe.read()
            self.assertRegex(
                report,
                "^== ExecutablePath =================================\n"
                "/bin/bash\n\n"
                "== ProblemType =================================\n"
                "Crash\n\n"
                "== Architecture =================================\n"
                "[^\n]+\n\n"
                "== CompressedFile =================================\n"
                "[^\n]+\n\n"
                "== CoreDump =================================\n"
                "[^\n]+\n\n"
                "== Date =================================\n"
                "[^\n]+\n\n"
                "== DistroRelease =================================\n"
                "[^\n]+\n\n"
                "== LongString =================================\n"
                "[^\n]+1042003[^\n]+\n\n"
                "== Signal =================================\n"
                "11\n\n"
                "== Uname =================================\n"
                "[^\n]+\n\n$",
            )

    @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_save_report_in_temp_directory(self, stdout_mock: io.StringIO) -> None:
        self.app.report["Package"] = "bash"
        with unittest.mock.patch.object(apport_cli.CLIDialog, "run") as run_mock:
            run_mock.return_value = 4
            self.app.ui_present_report_details()
        self.assertIn(_("Problem report file:"), stdout_mock.getvalue())



# Copyright (C) 2023 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for apport.sandboxutils."""

import os
import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

from apport.package_info import PackageInfo
from apport.report import Report
from apport.sandboxutils import _move_base_files_first, make_sandbox


class TestSandboxutils(unittest.TestCase):
    """Unit tests for apport.sandboxutils."""

    @staticmethod
    def _get_sample_report() -> Report:
        report = Report()
        report["Architecture"] = "amd64"
        report["DistroRelease"] = "Ubuntu 22.04"
        report["ExecutablePath"] = "/bin/bash"
        report["Signal"] = "11"
        return report

    @unittest.mock.patch("apport.sandboxutils.packaging", spec=PackageInfo)
    def test_make_sandbox(self, packaging_mock: MagicMock) -> None:
        """make_sandbox() for a sample report."""
        packaging_mock.install_packages.return_value = "obsolete\n"
        report = self._get_sample_report()
        sandbox, cache, outdated_msg = make_sandbox(report, "system")
        self.assertTrue(os.path.exists(sandbox), f"'{sandbox}' does not exist")
        self.assertTrue(os.path.exists(cache), f"'{cache}' does not exist")
        self.assertEqual(outdated_msg, "obsolete\nobsolete\n")
        self.assertEqual(packaging_mock.install_packages.call_count, 2)

    @unittest.mock.patch("apport.sandboxutils.packaging", spec=PackageInfo)
    def test_make_sandbox_install_packages_failure(
        self, packaging_mock: MagicMock
    ) -> None:
        """make_sandbox() where packaging.install_packages fails."""
        packaging_mock.install_packages.side_effect = SystemError("100% fail")
        with self.assertRaises(SystemExit):
            make_sandbox(self._get_sample_report(), "system")
        packaging_mock.install_packages.assert_called_once()

    @unittest.mock.patch("apport.sandboxutils.packaging", spec=PackageInfo)
    def test_make_sandbox_with_sandbox_dir(self, packaging_mock: MagicMock) -> None:
        """make_sandbox() with sandbox_dir set."""
        packaging_mock.install_packages.return_value = "obsolete\n"
        with tempfile.TemporaryDirectory(dir="/var/tmp") as tmpdir:
            config_dir = os.path.join(tmpdir, "config")
            cache_dir = os.path.join(tmpdir, "cache")
            sandbox_dir = os.path.join(tmpdir, "sandbox")
            sandbox, cache, outdated_msg = make_sandbox(
                self._get_sample_report(),
                config_dir,
                cache_dir=cache_dir,
                sandbox_dir=sandbox_dir,
            )
            self.assertEqual(sandbox, sandbox_dir)
            self.assertEqual(cache, cache_dir)
        self.assertEqual(outdated_msg, "obsolete\nobsolete\n")
        self.assertEqual(packaging_mock.install_packages.call_count, 2)

    def test_move_base_files_first_existing(self) -> None:
        """_move_base_files_first() with base-files in list."""
        pkgs: list[tuple[str, None | str]] = [
            ("chaos-marmosets", "0.1.2-2"),
            ("base-files", "13ubuntu9"),
            ("libc6", "2.39-0ubuntu8.2"),
        ]
        _move_base_files_first(pkgs)
        self.assertEqual(
            pkgs,
            [
                ("base-files", "13ubuntu9"),
                ("chaos-marmosets", "0.1.2-2"),
                ("libc6", "2.39-0ubuntu8.2"),
            ],
        )

    def test_move_base_files_first_missing(self) -> None:
        """_move_base_files_first() without base-files in list."""
        pkgs: list[tuple[str, None | str]] = [
            ("chaos-marmosets", "0.1.2-2"),
            ("libc6", "2.39-0ubuntu8.2"),
        ]
        _move_base_files_first(pkgs)
        self.assertEqual(
            pkgs, [("chaos-marmosets", "0.1.2-2"), ("libc6", "2.39-0ubuntu8.2")]
        )



# Copyright (C) 2025 Canonical Ltd.
# Author: Benjamin Drung <benjamin.drung@canonical.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Unit tests for apport.crash_impl.debian"""

from apport.crashdb_impl.debian import CrashDatabase
from apport.report import Report


def test_missing_sender() -> None:
    """Test missing sender in CrashDB configuration."""
    crashdb = CrashDatabase(None, {})
    report = Report()
    assert crashdb.accepts(report)
    assert report["UnreportableReason"] == (
        "Please configure sender settings in /etc/apport/crashdb.conf"
    )



"""Unit tests for apport-retrace."""

import io
import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

from tests.helper import import_module_from_file
from tests.paths import get_bin_directory

apport_retrace = import_module_from_file(get_bin_directory() / "apport-retrace")


@unittest.mock.patch.object(apport_retrace, "get_crashdb")
def test_malformed_crash_report(get_crashdb_mock: MagicMock) -> None:
    """Test apport-retrace to fail on malformed crash report."""
    with (
        tempfile.NamedTemporaryFile(mode="w+", suffix=".crash") as crash_file,
        unittest.mock.patch("sys.stderr", new_callable=io.StringIO) as stderr,
    ):
        crash_file.write(
            "ProblemType: Crash\nArchitecture: amd64\nPackage: gedit 46.2-2\n"
        )
        crash_file.flush()
        return_code = apport_retrace.main(["-x", "/usr/bin/gedit", crash_file.name])

    assert return_code == 2
    assert (
        stderr.getvalue()
        == "ERROR: report file does not contain one of the required fields:"
        " CoreDump DistroRelease\n"
    )
    get_crashdb_mock.assert_called_once_with(None)


@unittest.mock.patch.object(apport_retrace, "get_crashdb")
def test_malformed_kernel_crash_report(get_crashdb_mock: MagicMock) -> None:
    """Test apport-retrace to fail on malformed kernel crash report."""
    with (
        tempfile.NamedTemporaryFile(mode="w+", suffix=".crash") as crash_file,
        unittest.mock.patch("sys.stderr", new_callable=io.StringIO) as stderr,
    ):
        crash_file.write("ProblemType: KernelCrash\n")
        crash_file.flush()
        return_code = apport_retrace.main([crash_file.name])

    assert return_code == 2
    assert (
        stderr.getvalue() == "ERROR: report file does not contain the required fields\n"
    )
    get_crashdb_mock.assert_called_once_with(None)



"""Integration tests for the apport.hookutils module."""

import locale
import os
import re
import shutil
import subprocess
import sys
import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

import apport.hookutils
import apport.report
import problem_report
from tests.helper import restore_os_environ, skip_if_command_is_missing
from tests.paths import local_test_environment


class T(unittest.TestCase):
    """Integration tests for the apport.hookutils module."""

    orig_environ: dict[str, str]

    @classmethod
    def setUpClass(cls) -> None:
        cls.orig_environ = os.environ.copy()
        os.environ |= local_test_environment()

    @classmethod
    def tearDownClass(cls) -> None:
        os.environ.clear()
        os.environ.update(cls.orig_environ)

    def setUp(self) -> None:
        self.workdir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.workdir)

    @unittest.mock.patch(
        "apport.hookutils._root_command_prefix", MagicMock(return_value=[])
    )
    def test_attach_root_command_outputs(self) -> None:
        """attach_root_command_outputs()

        To not require root for this test, execute only commands that do
        not need root and mock _root_command_prefix().
        """
        report = problem_report.ProblemReport()
        command_map = {
            "BinaryBlob": r"printf '\303\050'",
            "EchoFoo": "echo foo",
            "EchoNothing": "echo",
            "EchoUnicode": "echo 'hä?'",
        }

        apport.hookutils.attach_root_command_outputs(report, command_map)

        self.assertEqual(report["BinaryBlob"], b"\303\050")
        self.assertEqual(report["EchoFoo"], "foo")
        self.assertEqual(report["EchoUnicode"], "hä?")
        self.assertEqual(
            set(report.keys()),
            {"Date", "BinaryBlob", "EchoFoo", "EchoUnicode", "ProblemType"},
        )

    @skip_if_command_is_missing("/usr/bin/as")
    def test_module_license_evaluation(self) -> None:
        """Module licenses can be validated correctly."""
        # pylint: disable=protected-access

        def _build_ko(license_name: str) -> str:
            ko_filename = os.path.join(self.workdir, f"{license_name}.ko")
            with tempfile.NamedTemporaryFile(
                mode="w+", prefix=f"{license_name}-", suffix=".S"
            ) as asm:
                asm.write(f'.section .modinfo\n.string "license={license_name}"\n')
                asm.flush()
                subprocess.check_call(["/usr/bin/as", asm.name, "-o", ko_filename])
            return ko_filename

        good_ko = _build_ko("GPL")
        bad_ko = _build_ko("BAD")

        # test:
        #  - unfindable module
        #  - fake GPL module
        #  - fake BAD module

        # direct license check
        self.assertEqual(
            apport.hookutils._get_module_license("does-not-exist"), "invalid"
        )
        self.assertIn("GPL", apport.hookutils._get_module_license(good_ko))
        self.assertIn("BAD", apport.hookutils._get_module_license(bad_ko))

        # check via nonfree_kernel_modules logic
        with tempfile.NamedTemporaryFile(mode="w+") as temp:
            temp.write(f"isofs\ndoes-not-exist\n{good_ko}\n{bad_ko}\n")
            temp.flush()
            nonfree = apport.hookutils.nonfree_kernel_modules(temp.name)
        self.assertIn("does-not-exist", nonfree)
        self.assertNotIn(good_ko, nonfree)
        self.assertIn(bad_ko, nonfree)

    @skip_if_command_is_missing("/sbin/modinfo")
    def test_real_module_license_evaluation(self) -> None:
        """Module licenses can be validated correctly for real module."""
        # pylint: disable=protected-access
        isofs_license = apport.hookutils._get_module_license("isofs")
        if isofs_license == "invalid":
            self.skipTest("kernel module 'isofs' not available")

        self.assertIn("GPL", isofs_license)

        with tempfile.NamedTemporaryFile() as temp:
            temp.write(b"isofs\n")
            temp.flush()
            nonfree = apport.hookutils.nonfree_kernel_modules(temp.name)
        self.assertNotIn("isofs", nonfree)

    def test_attach_file(self) -> None:
        """attach_file()"""
        with open("/etc/passwd", encoding="utf-8") as f:
            passwd_contents = f.read().strip()
        with open("/etc/issue", encoding="utf-8") as f:
            issue_contents = f.read().strip()

        # default key name
        report = problem_report.ProblemReport()
        default_keys = set(report)
        apport.hookutils.attach_file(report, "/etc/passwd")
        self.assertEqual(set(report) - default_keys, {".etc.passwd"})
        self.assertEqual(report[".etc.passwd"], passwd_contents)

        # custom key name
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file(report, "/etc/passwd", "Passwd")
        self.assertEqual(set(report) - default_keys, {"Passwd"})
        self.assertEqual(report["Passwd"], passwd_contents)

        # nonexisting file
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file(report, "/nonexisting")
        self.assertEqual(set(report) - default_keys, {".nonexisting"})
        self.assertTrue(report[".nonexisting"].startswith("Error: "))

        # symlink
        link = os.path.join(self.workdir, "symlink")
        os.symlink("/etc/passwd", link)
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file(report, link, "Symlink")
        self.assertEqual(set(report) - default_keys, {"Symlink"})
        self.assertTrue(report["Symlink"].startswith("Error: "))

        # directory symlink
        link = os.path.join(self.workdir, "dirsymlink")
        os.symlink("/etc", link)
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file(report, os.path.join(link, "passwd"), "DirSymlink")
        self.assertEqual(set(report) - default_keys, {"DirSymlink"})
        self.assertTrue(report["DirSymlink"].startswith("Error: "))

        # directory traversal
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file(report, "/etc/../etc/passwd", "Traversal")
        self.assertEqual(set(report) - default_keys, {"Traversal"})
        self.assertTrue(report["Traversal"].startswith("Error: "))

        # existing key
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file(report, "/etc/passwd")
        apport.hookutils.attach_file(report, "/etc/passwd")
        self.assertEqual(set(report) - default_keys, {".etc.passwd"})
        self.assertEqual(report[".etc.passwd"], passwd_contents)

        apport.hookutils.attach_file(
            report, "/etc/issue", ".etc.passwd", overwrite=False
        )
        self.assertEqual(set(report) - default_keys, {".etc.passwd", ".etc.passwd_"})
        self.assertEqual(report[".etc.passwd"], passwd_contents)
        self.assertEqual(report[".etc.passwd_"], issue_contents)

    def test_attach_file_binary(self) -> None:
        """attach_file() for binary files"""
        myfile = os.path.join(self.workdir, "data")
        with open(myfile, "wb") as f:
            f.write(b"a\xc3\xb6b\xffx")

        report = problem_report.ProblemReport()
        apport.hookutils.attach_file(report, myfile, key="data")
        self.assertEqual(report["data"], b"a\xc3\xb6b\xffx")

        apport.hookutils.attach_file(report, myfile, key="data", force_unicode=True)
        self.assertEqual(report["data"], b"a\xc3\xb6b\xef\xbf\xbdx".decode("UTF-8"))

    def test_attach_file_if_exists(self) -> None:
        """attach_file_if_exists()"""
        with open("/etc/passwd", encoding="utf-8") as f:
            passwd_contents = f.read().strip()

        # default key name
        report = problem_report.ProblemReport()
        default_keys = set(report)
        apport.hookutils.attach_file_if_exists(report, "/etc/passwd")
        self.assertEqual(set(report) - default_keys, {".etc.passwd"})
        self.assertEqual(report[".etc.passwd"], passwd_contents)

        # custom key name
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file_if_exists(report, "/etc/passwd", "Passwd")
        self.assertEqual(set(report) - default_keys, {"Passwd"})
        self.assertEqual(report["Passwd"], passwd_contents)

        # symlink
        link = os.path.join(self.workdir, "symlink")
        os.symlink("/etc/passwd", link)
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file_if_exists(report, link, "Symlink")
        self.assertEqual(set(report) - default_keys, {"Symlink"})
        self.assertTrue(report["Symlink"].startswith("Error: "))

        # nonexisting file
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file_if_exists(report, "/nonexisting")
        self.assertEqual(set(report) - default_keys, set())

        # directory traversal
        report = problem_report.ProblemReport()
        apport.hookutils.attach_file_if_exists(report, "/etc/../etc/passwd")
        self.assertEqual(set(report) - default_keys, set())

    def test_recent_syslog(self) -> None:
        """recent_syslog"""
        self.assertEqual(
            apport.hookutils.recent_syslog(re.compile("."), path="/nonexisting"), ""
        )
        self.assertEqual(
            apport.hookutils.recent_syslog(re.compile("ThisCantPossiblyHitAnything")),
            "",
        )
        if os.path.exists("/run/systemd/system") or os.access(
            "/var/log/syslog", os.R_OK
        ):
            self.assertNotEqual(len(apport.hookutils.recent_syslog(re.compile("."))), 0)

    def test_recent_syslog_overflow(self) -> None:
        """recent_syslog on a huge file"""
        log = os.path.join(self.workdir, "syslog")
        with open(log, "w", encoding="utf-8") as f:
            lines = 1000000
            while lines >= 0:
                f.write("Apr 20 11:30:00 komputer kernel: bogus message\n")
                lines -= 1

        mem_before = self._get_mem_usage()
        data = apport.hookutils.recent_syslog(re.compile("kernel"), path=log)
        mem_after = self._get_mem_usage()
        delta_kb = mem_after - mem_before
        sys.stderr.write(f"[Δ {delta_kb} kB] ")
        self.assertLess(delta_kb, 5000)

        self.assertRegex(data, "^Apr 20 11:30:00 komputer kernel: bogus message\n")
        self.assertGreater(len(data), 100000)
        self.assertLess(len(data), 1000000)

    @unittest.skipIf(
        apport.hookutils.apport.hookutils.in_session_of_problem(apport.report.Report())
        is None,
        "no logind session",
    )
    def test_in_session_of_problem(self) -> None:
        """in_session_of_problem()"""
        report = apport.report.Report(date="Sat Jan  1 12:00:00 2011")
        self.assertFalse(apport.hookutils.in_session_of_problem(report))

        report = apport.report.Report(date="Mon Oct 10 21:06:03 2009")
        self.assertFalse(apport.hookutils.in_session_of_problem(report))

        report = apport.report.Report()
        self.assertTrue(apport.hookutils.in_session_of_problem(report))

        self.assertIsNone(apport.hookutils.in_session_of_problem({}))

        orig_ctime = locale.getlocale(locale.LC_TIME)
        try:
            locale.setlocale(locale.LC_TIME, "C")

            report = apport.report.Report(date="Sat Jan  1 12:00:00 2011")
            self.assertFalse(apport.hookutils.in_session_of_problem(report))

            report = apport.report.Report(date="Mon Oct 10 21:06:03 2009")
            self.assertFalse(apport.hookutils.in_session_of_problem(report))

            report = apport.report.Report(date="Tue Jan  1 12:00:00 2038")
            self.assertTrue(apport.hookutils.in_session_of_problem(report))
        finally:
            locale.setlocale(locale.LC_TIME, orig_ctime)

    @restore_os_environ()
    def test_xsession_errors(self) -> None:
        """xsession_errors()"""
        with open(
            os.path.join(self.workdir, ".xsession-errors"), "w", encoding="UTF-8"
        ) as f:
            f.write("""\
Loading profile from /etc/profile
gnome-session[1948]: WARNING: standard glib warning
EggSMClient-CRITICAL **: egg_sm_client_set_mode: standard glib assertion
24/02/2012 11:14:46 Sending credentials s3kr1t

** WARNING **: nonstandard warning

WARN  2012-02-24 11:23:47 unity <unknown>:0 some unicode ♥ ♪

GNOME_KEYRING_CONTROL=/tmp/keyring-u7hrD6

(gnome-settings-daemon:5115): Gdk-WARNING **: The program\
 'gnome-settings-daemon' received an X Window System error.
This probably reflects a bug in the program.
The error was 'BadMatch (invalid parameter attributes)'.
  (Details: serial 723 error_code 8 request_code 143 minor_code 22)
  (Note to programmers: normally, X errors are reported asynchronously;
   that is, you will receive the error a while after causing it.
   To debug your program, run it with the --sync command line
   option to change this behavior. You can then get a meaningful
   backtrace from your debugger if you break on the gdk_x_error() function.)"

GdkPixbuf-CRITICAL **: gdk_pixbuf_scale_simple: another standard glib assertion
""")
        os.environ["HOME"] = self.workdir

        # explicit pattern
        pattern = re.compile("notfound")
        self.assertEqual(apport.hookutils.xsession_errors(pattern), "")

        pattern = re.compile(r"^\w+-CRITICAL")
        res = apport.hookutils.xsession_errors(pattern).splitlines()
        self.assertEqual(len(res), 2)
        self.assertTrue(res[0].startswith("EggSMClient-CRITICAL"))
        self.assertTrue(res[1].startswith("GdkPixbuf-CRITICAL"))

        # default pattern includes glib assertions and X Errors
        res = apport.hookutils.xsession_errors()
        self.assertNotIn("nonstandard warning", res)
        self.assertNotIn("keyring", res)
        self.assertNotIn("credentials", res)
        self.assertIn("WARNING: standard glib warning", res)
        self.assertIn("GdkPixbuf-CRITICAL", res)
        self.assertIn("'gnome-settings-daemon' received an X Window", res)
        self.assertIn("BadMatch", res)
        self.assertIn("serial 723", res)

    @staticmethod
    @unittest.mock.patch(
        "apport.hookutils._root_command_prefix", MagicMock(return_value=[])
    )
    def test_no_crashes() -> None:
        """Functions do not crash (very shallow)."""
        report = problem_report.ProblemReport()
        apport.hookutils.attach_hardware(report)
        apport.hookutils.attach_alsa(report)
        apport.hookutils.attach_network(report)
        apport.hookutils.attach_wifi(report)
        apport.hookutils.attach_printing(report)
        apport.hookutils.attach_conffiles(report, "bash")
        apport.hookutils.attach_conffiles(report, "apport")
        apport.hookutils.attach_conffiles(report, "nonexisting")
        apport.hookutils.attach_default_grub(report)

    def test_command_output(self) -> None:
        """Test apport.hookutils.command_output."""
        with restore_os_environ():
            os.environ["LC_MESSAGES"] = "en_US.UTF-8"
            # default mode: disable translations
            out = apport.hookutils.command_output(["env"])
            self.assertIn("LC_MESSAGES=C", out)

            # keep locale
            out = apport.hookutils.command_output(["env"], keep_locale=True)
            self.assertNotIn("LC_MESSAGES=C", out)

        # nonexisting binary
        out = apport.hookutils.command_output(["/non existing"])
        self.assertTrue(out.startswith("Error: [Errno 2]"))

        # stdin
        out = apport.hookutils.command_output(["cat"], input=b"hello")
        self.assertEqual(out, "hello")

    @staticmethod
    def _get_mem_usage() -> int:
        """Get current memory usage in kB."""
        with open("/proc/self/status", encoding="utf-8") as f:
            for line in f:
                if line.startswith("VmSize:"):
                    return int(line.split()[1])

        raise SystemError("did not find VmSize: in /proc/self/status")



# Copyright (C) 2026 Canonical Ltd.
# Author: Benjamin Drung <bdrung@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Functions to operate on files in /proc."""

import dataclasses


@dataclasses.dataclass
class Meminfo:
    """Data from /proc/meminfo. Values are in KiB.

    Only the needed values from /proc/meminfo are provided.
    This dataclass can be extended when needed.
    """

    mem_total: int
    mem_free: int
    mem_available: int
    cached: int
    writeback: int


def parse_meminfo() -> Meminfo:
    """Parse /proc/meminfo and return a dictionary for the requested keys."""
    remaining_keys = {"MemTotal", "MemFree", "MemAvailable", "Cached", "Writeback"}
    meminfo = {}
    with open("/proc/meminfo", "r", encoding="utf-8") as meminfo_file:
        for line in meminfo_file:
            key, remaining = line.split(":", maxsplit=1)
            if key not in remaining_keys:
                continue
            value, unit = remaining.strip().split(maxsplit=2)
            assert unit == "kB" or value == "0"
            meminfo[key] = int(value)
            remaining_keys.remove(key)
            if not remaining_keys:
                return Meminfo(
                    meminfo["MemTotal"],
                    meminfo["MemFree"],
                    meminfo["MemAvailable"],
                    meminfo["Cached"],
                    meminfo["Writeback"],
                )
    raise KeyError(f"{', '.join(sorted(remaining_keys))} not found in /proc/meminfo")



"""Apport Python module."""

# for faster module loading and avoiding circular dependencies
# pylint: disable=import-outside-toplevel

import typing

# Import apport.packaging to shadow it afterwards.
import apport.packaging as _  # noqa: F401

__all__ = [
    "Report",
    "error",
    "fatal",
    "log",
    "memdbg",
    "packaging",
    "unicode_gettext",
    "warning",
]

if typing.TYPE_CHECKING:  # pragma: no cover
    from apport.packaging_impl import impl as packaging
    from apport.report import Report
else:
    # wraps an object; pylint: disable-next=invalid-name
    def Report(*args, **kwargs):
        """Lazy loading of apport.report.Report()."""
        from apport.report import Report as cls

        return cls(*args, **kwargs)

    # wrapper object; pylint: disable-next=too-few-public-methods
    class _LazyLoadingPackaging:
        def __getattribute__(self, name):
            # The packaging object will be replaced by the imported object.
            # pylint: disable-next=global-statement
            global packaging
            from apport.packaging_impl import impl as packaging

            return packaging.__getattribute__(name)

    packaging = _LazyLoadingPackaging()


def _logging_function(function_name):
    def _wrapped_logging_function(*args, **kwargs):
        import warnings

        import apport.logging

        warnings.warn(
            f"apport.{function_name}() is deprecated."
            f" Please use apport.logging.{function_name}() directly instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return getattr(apport.logging, function_name)(*args, **kwargs)

    return _wrapped_logging_function


error = _logging_function("error")
fatal = _logging_function("fatal")
log = _logging_function("log")
memdbg = _logging_function("memdbg")
warning = _logging_function("warning")


def unicode_gettext(message):
    """Return the localized translation of message."""
    import gettext
    import warnings

    warnings.warn(
        "apport.unicode_gettext() is deprecated."
        " Please use gettext.gettext() directly instead.",
        PendingDeprecationWarning,
        stacklevel=2,
    )
    return gettext.gettext(message)



# Copyright (C) 2024 Canonical Ltd.
# Author: Benjamin Drung <bdrung@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Validate XML files."""

import subprocess

import pytest

from tests.paths import is_local_source_directory


def get_policy_xml() -> str:
    """Determine path to Apport PolicyKit XML."""
    if not is_local_source_directory():
        return "/usr/share/polkit-1/actions/com.ubuntu.apport.policy"
    policy_xml = "build/share/polkit-1/actions/com.ubuntu.apport.policy"
    return policy_xml


def test_validate_xml() -> None:
    """Validate Apport PolicyKit XML."""
    cmd = [
        "xmllint",
        "--noout",
        "--nonet",
        "--dtdvalid",
        "/usr/share/polkit-1/policyconfig-1.dtd",
        get_policy_xml(),
    ]
    try:
        subprocess.check_call(cmd)
    except FileNotFoundError as error:
        pytest.skip(f"{error.filename} not available")



# Copyright (C) 2007 - 2011 Canonical Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
# the full text of the license.

"""Interactive functions which can be used in package hooks"""

import threading
import time


class HookUI:
    """Interactive functions which can be used in package hooks.

    This provides an interface for package hooks which need to ask interactive
    questions. Directly passing the UserInterface instance to the hooks needs
    to be avoided, since we need to call the UI methods in a different thread,
    and also don't want hooks to be able to poke in the UI.
    """

    def __init__(self, ui):
        """Create a HookUI object.

        ui is the UserInterface instance to wrap.
        """
        self.ui = ui

        # variables for communicating with the UI thread
        self._request_event = threading.Event()
        self._response_event = threading.Event()
        self._request_fn = None
        self._request_args = None
        self._response = None

    #
    # API for hooks
    #

    def information(self, text):
        """Show an information with OK/Cancel buttons.

        This can be used for asking the user to perform a particular action,
        such as plugging in a device which does not work.
        """
        return self._trigger_ui_request("ui_info_message", "", text)

    def yesno(self, text):
        """Show a yes/no question.

        Return True if the user selected "Yes", False if selected "No" or
        "None" on cancel/dialog closing.
        """
        return self._trigger_ui_request("ui_question_yesno", text)

    def choice(self, text, options, multiple=False):
        """Show an question with predefined choices.

        options is a list of strings to present. If multiple is True, they
        should be check boxes, if multiple is False they should be radio
        buttons.

        Return list of selected option indexes, or None if the user cancelled.
        If multiple == False, the list will always have one element.
        """
        return self._trigger_ui_request("ui_question_choice", text, options, multiple)

    def file(self, text):
        """Show a file selector dialog.

        Return path if the user selected a file, or None if cancelled.
        """
        return self._trigger_ui_request("ui_question_file", text)

    #
    # internal API for inter-thread communication
    #

    def _trigger_ui_request(self, fn, *args):
        """Called by HookUi functions in info collection thread."""
        # only one at a time
        assert not self._request_event.is_set()
        assert not self._response_event.is_set()
        assert self._request_fn is None

        self._response = None
        self._request_fn = fn
        self._request_args = args
        self._request_event.set()
        self._response_event.wait()

        self._request_fn = None
        self._response_event.clear()

        return self._response

    def process_event(self):
        """Called by GUI thread to check and process hook UI requests."""
        # sleep for 0.1 seconds to wait for events
        self._request_event.wait(0.1)
        if not self._request_event.is_set():
            return

        assert not self._response_event.is_set()
        self._request_event.clear()
        self._response = getattr(self.ui, self._request_fn)(*self._request_args)
        self._response_event.set()


class NoninteractiveHookUI(HookUI):
    """HookUI variant that does not ask the user any questions."""

    def __init__(self):
        super().__init__(None)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def information(self, text):
        return None

    def yesno(self, text):
        return None

    def choice(self, text, options, multiple=False):
        return None

    def file(self, text):
        return None

    def process_event(self):
        # Give other threads some chance to run
        time.sleep(0.1)



"""Unit tests for the apport.hookutils module."""

import datetime
import os
import re
import subprocess
import unittest
import unittest.mock
from unittest.mock import MagicMock, Mock

import apport.hookutils
import apport.report
from problem_report import ProblemReport

IW_REG_LIST_DE = b"""\
global
country 00: DFS-UNSET
	(755 - 928 @ 2), (N/A, 20), (N/A), PASSIVE-SCAN
	(2402 - 2472 @ 40), (N/A, 20), (N/A)
	(2457 - 2482 @ 20), (N/A, 20), (N/A), AUTO-BW, PASSIVE-SCAN
	(2474 - 2494 @ 20), (N/A, 20), (N/A), NO-OFDM, PASSIVE-SCAN
	(5170 - 5250 @ 80), (N/A, 20), (N/A), AUTO-BW, PASSIVE-SCAN
	(5250 - 5330 @ 80), (N/A, 20), (0 ms), DFS, AUTO-BW, PASSIVE-SCAN
	(5490 - 5730 @ 160), (N/A, 20), (0 ms), DFS, PASSIVE-SCAN
	(5735 - 5835 @ 80), (N/A, 20), (N/A), PASSIVE-SCAN
	(57240 - 63720 @ 2160), (N/A, 0), (N/A)

phy#0 (self-managed)
country DE: DFS-UNSET
	(2402 - 2437 @ 40), (6, 22), (N/A), AUTO-BW, NO-HT40MINUS, NO-80MHZ, NO-160MHZ
	(2422 - 2462 @ 40), (6, 22), (N/A), AUTO-BW, NO-80MHZ, NO-160MHZ
	(2447 - 2482 @ 40), (6, 22), (N/A), AUTO-BW, NO-HT40PLUS, NO-80MHZ, NO-160MHZ
	(5170 - 5190 @ 160), (6, 22), (N/A), NO-OUTDOOR, AUTO-BW, IR-CONCURRENT, NO-HT40MINUS, NO-320MHZ, PASSIVE-SCAN
	(5190 - 5210 @ 160), (6, 22), (N/A), NO-OUTDOOR, AUTO-BW, IR-CONCURRENT, NO-HT40PLUS, NO-320MHZ, PASSIVE-SCAN
	(5210 - 5230 @ 160), (6, 22), (N/A), NO-OUTDOOR, AUTO-BW, IR-CONCURRENT, NO-HT40MINUS, NO-320MHZ, PASSIVE-SCAN
	(5230 - 5250 @ 160), (6, 22), (N/A), NO-OUTDOOR, AUTO-BW, IR-CONCURRENT, NO-HT40PLUS, NO-320MHZ, PASSIVE-SCAN
	(5250 - 5270 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40MINUS, NO-320MHZ, PASSIVE-SCAN
	(5270 - 5290 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40PLUS, NO-320MHZ, PASSIVE-SCAN
	(5290 - 5310 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40MINUS, NO-320MHZ, PASSIVE-SCAN
	(5310 - 5330 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40PLUS, NO-320MHZ, PASSIVE-SCAN
	(5490 - 5510 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40MINUS, NO-320MHZ, PASSIVE-SCAN
	(5510 - 5530 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40PLUS, NO-320MHZ, PASSIVE-SCAN
	(5530 - 5550 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40MINUS, NO-320MHZ, PASSIVE-SCAN
	(5550 - 5570 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40PLUS, NO-320MHZ, PASSIVE-SCAN
	(5570 - 5590 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40MINUS, NO-320MHZ, PASSIVE-SCAN
	(5590 - 5610 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40PLUS, NO-320MHZ, PASSIVE-SCAN
	(5610 - 5630 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40MINUS, NO-320MHZ, PASSIVE-SCAN
	(5630 - 5650 @ 160), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40PLUS, NO-320MHZ, PASSIVE-SCAN
	(5650 - 5670 @ 80), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40MINUS, NO-160MHZ, NO-320MHZ, PASSIVE-SCAN
	(5670 - 5690 @ 80), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40PLUS, NO-160MHZ, NO-320MHZ, PASSIVE-SCAN
	(5690 - 5710 @ 80), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40MINUS, NO-160MHZ, NO-320MHZ, PASSIVE-SCAN
	(5710 - 5730 @ 80), (6, 22), (0 ms), DFS, AUTO-BW, NO-HT40PLUS, NO-160MHZ, NO-320MHZ, PASSIVE-SCAN
	(5735 - 5755 @ 80), (6, 22), (N/A), AUTO-BW, NO-HT40MINUS, NO-160MHZ, NO-320MHZ
	(5755 - 5775 @ 80), (6, 22), (N/A), AUTO-BW, NO-HT40PLUS, NO-160MHZ, NO-320MHZ
	(5775 - 5795 @ 80), (6, 22), (N/A), AUTO-BW, NO-HT40MINUS, NO-160MHZ, NO-320MHZ
	(5795 - 5815 @ 80), (6, 22), (N/A), AUTO-BW, NO-HT40PLUS, NO-160MHZ, NO-320MHZ
	(5815 - 5835 @ 20), (6, 22), (N/A), AUTO-BW, NO-HT40MINUS, NO-HT40PLUS, NO-80MHZ, NO-160MHZ, NO-320MHZ

"""  # noqa


class TestHookutils(unittest.TestCase):
    # pylint: disable=missing-function-docstring
    """Test apport.hookutils module."""

    maxDiff = None

    @unittest.mock.patch("os.path.isdir", MagicMock(return_value=True))
    @unittest.mock.patch("os.listdir")
    @unittest.mock.patch("os.stat")
    @unittest.mock.patch("apport.hookutils.read_file")
    def test_attach_dmi(
        self, read_file_mock: MagicMock, stat_mock: MagicMock, listdir_mock: MagicMock
    ) -> None:
        """attach_dmi()"""

        def stat(
            content: str = "private\n",
            st_mode: int = 0o100444,
            st_nlink: int = 1,
            st_size: int = 4096,
        ) -> tuple[str, os.stat_result]:
            st_time = 1698056204.0
            stat = os.stat_result(
                (st_mode, 1337, 22, st_nlink, 0, 0, st_size, st_time, st_time, st_time)
            )
            return (content, stat)

        dmi_files = {
            "bios_date": stat("07/20/2022\n"),
            "board_serial": stat(st_mode=0o100400),
            "uevent": stat("MODALIAS=dmi:[...]\n", st_mode=0o100644),
            "product_serial": stat(st_mode=0o100400),
            "product_name": stat("B550I AORUS PRO AX\n"),
            "sys_vendor": stat("Gigabyte Technology Co., Ltd.\n"),
            "power": stat(st_mode=0o40755, st_nlink=2, st_size=0),
            "bios_version": stat("F16e\n"),
            "bios_release": stat("5.17\n"),
            "board_vendor": stat("Gigabyte Technology Co., Ltd.\n"),
            "subsystem": stat(st_mode=0o40755, st_nlink=2, st_size=0),
            "product_family": stat("B550 MB\n"),
            "product_uuid": stat(
                "2cf4a3c2-8f76-4f38-95b5-77c8a9f6ec4f\n", st_mode=0o100400
            ),
            "bios_vendor": stat("American Megatrends International, LLC.\n"),
            "board_name": stat("B550I AORUS PRO AX\n"),
        }

        def mock_os_stat(path: str) -> os.stat_result:
            return dmi_files[os.path.basename(path)][1]

        def mock_read_file(path: str) -> str:
            return dmi_files[os.path.basename(path)][0].strip()

        listdir_mock.return_value = list(dmi_files.keys())
        stat_mock.side_effect = mock_os_stat
        read_file_mock.side_effect = mock_read_file
        report: dict[str, str] = {}
        apport.hookutils.attach_dmi(report)
        self.assertEqual(
            report,
            {
                "dmi.bios.date": "07/20/2022",
                "dmi.bios.release": "5.17",
                "dmi.bios.vendor": "American Megatrends International, LLC.",
                "dmi.bios.version": "F16e",
                "dmi.board.name": "B550I AORUS PRO AX",
                "dmi.board.vendor": "Gigabyte Technology Co., Ltd.",
                "dmi.product.family": "B550 MB",
                "dmi.product.name": "B550I AORUS PRO AX",
                "dmi.sys.vendor": "Gigabyte Technology Co., Ltd.",
                "MachineType": "Gigabyte Technology Co., Ltd. B550I AORUS PRO AX",
            },
        )

    @unittest.mock.patch("apport.hookutils.root_command_output")
    def test_attach_dmesg(self, root_command_output_mock: MagicMock) -> None:
        """attach_dmesg()"""
        root_command_output_mock.return_value = "[30804.972250] CPU0 is up"

        report: dict[str, str] = {}
        apport.hookutils.attach_dmesg(report)
        self.assertEqual(report, {"CurrentDmesg": "[30804.972250] CPU0 is up"})

        root_command_output_mock.assert_called_once_with(["dmesg"])

    @staticmethod
    @unittest.mock.patch("apport.hookutils.execute_multiple_root_commands")
    @unittest.mock.patch("os.path.exists")
    def test_attach_mac_events_no_profile(
        path_exists_mock: MagicMock, multiple_root_commands_mock: MagicMock
    ) -> None:
        """attach_mac_events() with default profile"""
        apparmor_status_line = """[ 3599.170011] audit: type=1326 audit(1733479659.940:2331): auid=1000 uid=1000 gid=1000 ses=3 subj=snap.chromium.chromium pid=26013 comm="chrome" exe="/snap/chromium/3002/usr/lib/chromium-browser/chrome" sig=0 arch=c000003e syscall=444 compat=0 ip=0x77858971e88d code=0x50000"""  # nopep8 pylint: disable=line-too-long
        apparmor_denied_line = """[ 3599.407888] audit: type=1400 audit(1733479660.178:2332): apparmor="DENIED" operation="open" class="file" profile="snap.chromium.chromium" name="/etc/vulkan/implicit_layer.d/" pid=26219 comm="chrome" requested_mask="r" denied_mask="r" fsuid=1000 ouid=0"""  # nopep8 pylint: disable=line-too-long
        bogus_line = """[30804.972250] CPU0 is up"""
        # TODO: create fake audit logs and version_signature files
        path_exists_mock.return_value = False
        multiple_root_commands_mock.return_value = {
            "dmesg": f"{apparmor_status_line}\n{apparmor_denied_line}\n{bogus_line}"
        }

        report = ProblemReport()
        apport.hookutils.attach_mac_events(report, profiles=None)

        assert multiple_root_commands_mock.called
        assert path_exists_mock.called

        assert apparmor_denied_line[32:] in report["KernLog"]
        assert apparmor_status_line[32:] in report["KernLog"]
        assert "CPU0 is up" not in report["KernLog"]

        assert "apparmor" in report.get_tags()

    def test_anonymized_udevdb(self) -> None:
        # One device entry with an ext4 filesystem named "john-doe-rootfs"
        # Many properties have been removed to make the test case cleaner
        udevdb = """
P: /devices/pci0000:00/0000:00:04.0/virtio2/block/vda/vda1
S: disk/by-path/pci-0000:00:04.0-part/by-label/john-doe-rootfs
S: disk/by-label/john-doe-rootfs
E: DEVPATH=/devices/pci0000:00/0000:00:04.0/virtio2/block/vda/vda1
E: DEVNAME=/dev/vda1
E: ID_FS_LABEL=john-doe-rootfs
E: ID_FS_LABEL_ENC=john-doe-rootfs
E: DEVLINKS=/dev/disk/by-path/pci-0000:00:04.0-part/by-partuuid/2722bff2-40ac-47e6-8025-5b7889adb69a /dev/disk/by-path/pci-0000:00:04.0-part1 /dev/disk/by-path/pci-0000:00:04.0-part/by-partnum/1 /dev/disk/by-uuid/e6134f84-e3eb-4ae4-8c43-959b9740bd3e /dev/disk/by-path/pci-0000:00:04.0-part/by-uuid/e6134f84-e3eb-4ae4-8c43-959b9740bd3e /dev/disk/by-path/virtio-pci-0000:00:04.0-part1 /dev/disk/by-path/pci-0000:00:04.0-part/by-label/john-doe-rootfs /dev/disk/by-label/john-doe-rootfs /dev/disk/by-diskseq/9-part1 /dev/disk/by-partuuid/2722bff2-40ac-47e6-8025-5b7889adb69a
"""  # nopep8
        expected = """
P: /devices/pci0000:00/0000:00:04.0/virtio2/block/vda/vda1
S: disk/by-path/pci-0000:00:04.0-part/by-label/<hidden>
S: disk/by-label/<hidden>
E: DEVPATH=/devices/pci0000:00/0000:00:04.0/virtio2/block/vda/vda1
E: DEVNAME=/dev/vda1
E: ID_FS_LABEL=<hidden>
E: ID_FS_LABEL_ENC=<hidden>
E: DEVLINKS=/dev/disk/by-path/pci-0000:00:04.0-part/by-partuuid/2722bff2-40ac-47e6-8025-5b7889adb69a /dev/disk/by-path/pci-0000:00:04.0-part1 /dev/disk/by-path/pci-0000:00:04.0-part/by-partnum/1 /dev/disk/by-uuid/e6134f84-e3eb-4ae4-8c43-959b9740bd3e /dev/disk/by-path/pci-0000:00:04.0-part/by-uuid/e6134f84-e3eb-4ae4-8c43-959b9740bd3e /dev/disk/by-path/virtio-pci-0000:00:04.0-part1 /dev/disk/by-path/pci-0000:00:04.0-part/by-label/<hidden> /dev/disk/by-label/<hidden> /dev/disk/by-diskseq/9-part1 /dev/disk/by-partuuid/2722bff2-40ac-47e6-8025-5b7889adb69a
"""  # nopep8

        self.assertEqual(expected, apport.hookutils.anonymized_udevdb(udevdb))

    @staticmethod
    @unittest.mock.patch("apport.hookutils.execute_multiple_root_commands")
    @unittest.mock.patch("os.path.exists")
    def test_attach_mac_events_good_profile(
        path_exists_mock: MagicMock, multiple_root_commands_mock: MagicMock
    ) -> None:
        """attach_mac_events() with a matching profile"""
        apparmor_status_line = """[ 3599.170011] audit: type=1326 audit(1733479659.940:2331): auid=1000 uid=1000 gid=1000 ses=3 subj=snap.chromium.chromium pid=26013 comm="chrome" exe="/snap/chromium/3002/usr/lib/chromium-browser/chrome" sig=0 arch=c000003e syscall=444 compat=0 ip=0x77858971e88d code=0x50000"""  # nopep8 pylint: disable=line-too-long
        apparmor_denied_line = """[ 3599.407888] audit: type=1400 audit(1733479660.178:2332): apparmor="DENIED" operation="open" class="file" profile="snap.chromium.chromium" name="/etc/vulkan/implicit_layer.d/" pid=26219 comm="chrome" requested_mask="r" denied_mask="r" fsuid=1000 ouid=0"""  # nopep8 pylint: disable=line-too-long
        bogus_line = """[30804.972250] CPU0 is up"""
        # TODO: create fake audit logs and version_signature files
        path_exists_mock.return_value = False
        multiple_root_commands_mock.return_value = {
            "dmesg": f"{apparmor_status_line}\n{apparmor_denied_line}\n{bogus_line}"
        }

        report = ProblemReport()
        apport.hookutils.attach_mac_events(report, profiles=["snap.chromium.chromium"])

        assert multiple_root_commands_mock.called
        assert path_exists_mock.called

        assert apparmor_denied_line[32:] in report["KernLog"]
        assert apparmor_status_line[32:] in report["KernLog"]
        assert "CPU0 is up" not in report["KernLog"]

        assert "apparmor" in report.get_tags()

    @staticmethod
    @unittest.mock.patch("apport.hookutils.execute_multiple_root_commands")
    @unittest.mock.patch("os.path.exists")
    def test_attach_mac_events_bad_profile(
        path_exists_mock: MagicMock, multiple_root_commands_mock: MagicMock
    ) -> None:
        """attach_mac_events() with a bad profile"""
        apparmor_status_line = """[ 3599.170011] audit: type=1326 audit(1733479659.940:2331): auid=1000 uid=1000 gid=1000 ses=3 subj=snap.chromium.chromium pid=26013 comm="chrome" exe="/snap/chromium/3002/usr/lib/chromium-browser/chrome" sig=0 arch=c000003e syscall=444 compat=0 ip=0x77858971e88d code=0x50000"""  # nopep8 pylint: disable=line-too-long
        apparmor_denied_line = """[ 3599.407888] audit: type=1400 audit(1733479660.178:2332): apparmor="DENIED" operation="open" class="file" profile="snap.chromium.chromium" name="/etc/vulkan/implicit_layer.d/" pid=26219 comm="chrome" requested_mask="r" denied_mask="r" fsuid=1000 ouid=0"""  # nopep8 pylint: disable=line-too-long
        bogus_line = """[30804.972250] CPU0 is up"""
        # TODO: create fake audit logs and version_signature files
        path_exists_mock.return_value = False
        multiple_root_commands_mock.return_value = {
            "dmesg": f"{apparmor_status_line}\n{apparmor_denied_line}\n{bogus_line}"
        }

        report = ProblemReport()
        apport.hookutils.attach_mac_events(report, profiles=["snap.firefox"])

        assert multiple_root_commands_mock.called
        assert path_exists_mock.called

        assert apparmor_denied_line[32:] in report["KernLog"]
        assert apparmor_status_line[32:] in report["KernLog"]
        assert "CPU0 is up" not in report["KernLog"]

        assert "apparmor" not in report.get_tags()

    def test_dmesg_overwrite(self) -> None:
        """attach_dmesg() does not overwrite already existing data"""
        report = {"CurrentDmesg": "existingcurrent"}

        apport.hookutils.attach_dmesg(report)
        self.assertEqual(report["CurrentDmesg"], "existingcurrent")

    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch("os.path.exists", MagicMock(return_value=True))
    def test_attach_journal_errors_with_date(self, run_mock: MagicMock) -> None:
        run_mock.return_value = subprocess.CompletedProcess(
            args=MagicMock(), returncode=0, stdout=b"journalctl output", stderr=b""
        )
        now = datetime.datetime.now()

        report = apport.report.Report(date=now.strftime("%a %b %d %H:%M:%S %Y"))
        apport.hookutils.attach_journal_errors(report)

        self.assertEqual(run_mock.call_count, 1)
        self.assertEqual(report.get("JournalErrors"), "journalctl output")
        self.assertEqual(
            run_mock.call_args[0][0],
            [
                "journalctl",
                "--priority=warning",
                f"--since=@{int(now.timestamp()) - 10}",
                f"--until=@{int(now.timestamp()) + 10}",
            ],
        )

    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch("os.path.exists", MagicMock(return_value=True))
    def test_attach_journal_errors_without_date(self, run_mock: MagicMock) -> None:
        run_mock.return_value = subprocess.CompletedProcess(
            args=MagicMock(), returncode=0, stdout=b"journalctl output", stderr=b""
        )

        report = apport.report.Report()
        del report["Date"]
        apport.hookutils.attach_journal_errors(report)

        self.assertEqual(run_mock.call_count, 1)
        self.assertEqual(report.get("JournalErrors"), "journalctl output")
        self.assertEqual(
            run_mock.call_args[0][0],
            ["journalctl", "--priority=warning", "-b", "--lines=1000"],
        )

    @unittest.mock.patch("apport.hookutils.execute_multiple_root_commands")
    def test_attach_mac_events_apparmor(
        self, execute_multiple_root_commands_mock: MagicMock
    ) -> None:
        # TODO: Split into separate test cases
        # pylint: disable=too-many-statements
        """Test apparmor tag calculation of attach_mac_events()"""
        execute_multiple_root_commands_mock.return_value = {}
        denied_log = (
            "[  351.624338] type=1400 audit(1343775571.688:27):"
            ' apparmor="DENIED" operation="capable" parent=1'
            ' profile="/usr/sbin/cupsd" pid=1361 comm="cupsd" pid=1361'
            ' comm="cupsd" capability=36  capname="block_suspend"\n'
        )

        denied_hex = (
            "[  351.624338] type=1400 audit(1343775571.688:27):"
            ' apparmor="DENIED" operation="capable" parent=1'
            ' profile=2F7573722F7362696E2F6375707364 pid=1361 comm="cupsd"'
            ' pid=1361 comm="cupsd" capability=36  capname="block_suspend"\n'
        )

        # No AppArmor messages
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = (
            "[    2.997534] i915 0000:00:02.0:"
            " power state changed by ACPI to D0\n"
            "[    2.997541] i915 0000:00:02.0:"
            " PCI INT A -> GSI 16 (level, low)\n"
            "[    2.997544] i915 0000:00:02.0: setting latency timer to 64\n"
            "[    3.061584] i915 0000:00:02.0: irq 42 for MSI/MSI-X\n"
        )

        apport.hookutils.attach_mac_events(report)
        self.assertNotIn("Tags", report)
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor message, but not a denial
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = (
            "[   32.420248] type=1400 audit(1344562672.449:2):"
            ' apparmor="STATUS" operation="profile_load" name="/sbin/dhclient"'
            ' pid=894 comm="apparmor_parser"\n'
        )

        apport.hookutils.attach_mac_events(report)
        self.assertNotIn("Tags", report)
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial, empty tags, no profile specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report)
        self.assertEqual(report["Tags"], "apparmor")
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor hex-encoded denial, no profile specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_hex

        apport.hookutils.attach_mac_events(report)
        self.assertEqual(report["Tags"], "apparmor")
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial in AuditLog
        report = ProblemReport()
        report["AuditLog"] = denied_log
        report["KernLog"] = "some dmesg log"

        apport.hookutils.attach_mac_events(report)
        self.assertEqual(report["Tags"], "apparmor")
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial, pre-existing tags, no profile specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_log
        report["Tags"] = "bogustag"

        apport.hookutils.attach_mac_events(report)
        self.assertEqual(report["Tags"], "apparmor bogustag")
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial, single profile specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cupsd")
        self.assertEqual(report["Tags"], "apparmor")
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial, regex profile specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cups.*")
        self.assertEqual(report["Tags"], "apparmor")
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial, subset profile specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cup")
        self.assertNotIn("Tags", report)
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor hex-encoded denial, single profile specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_hex

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cupsd")
        self.assertEqual(report["Tags"], "apparmor")
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial, single different profile specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_log

        apport.hookutils.attach_mac_events(report, "/usr/sbin/nonexistent")
        self.assertNotIn("Tags", report)
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial, multiple profiles specified
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_log
        profiles = ["/usr/bin/nonexistent", "/usr/sbin/cupsd"]

        apport.hookutils.attach_mac_events(report, profiles)
        self.assertEqual(report["Tags"], "apparmor")
        execute_multiple_root_commands_mock.assert_called_with({})

        # AppArmor denial, multiple different profiles
        report = ProblemReport()
        report["AuditLog"] = "some audit log"
        report["KernLog"] = denied_log
        profiles = ["/usr/bin/nonexistent", "/usr/sbin/anotherone"]

        apport.hookutils.attach_mac_events(report, profiles)
        self.assertNotIn("Tags", report)
        execute_multiple_root_commands_mock.assert_called_with({})

        # Multiple AppArmor denials, second match
        report = ProblemReport()
        report["KernLog"] = (
            "[  351.624338] type=1400 audit(1343775571.688:27):"
            ' apparmor="DENIED" operation="capable" parent=1'
            ' profile="/usr/sbin/blah" pid=1361 comm="cupsd" pid=1361'
            ' comm="cupsd" capability=36  capname="block_suspend"\n'
            "[  351.624338] type=1400 audit(1343775571.688:27):"
            ' apparmor="DENIED" operation="capable" parent=1'
            ' profile="/usr/sbin/cupsd" pid=1361 comm="cupsd" pid=1361'
            ' comm="cupsd" capability=36  capname="block_suspend"\n'
        )

        apport.hookutils.attach_mac_events(report, "/usr/sbin/cupsd")
        self.assertEqual(report["Tags"], "apparmor")
        execute_multiple_root_commands_mock.assert_called_with({})

    @unittest.mock.patch("glob.glob")
    @unittest.mock.patch("os.path.exists")
    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch(
        "apport.hookutils.recent_syslog", MagicMock(return_value="some recent logs")
    )
    def test_attach_wifi(
        self, run_mock: MagicMock, exists_mock: MagicMock, glob_mock: MagicMock
    ) -> None:
        """Test attach_wifi() with some real world data from Ubuntu 25.04.

        This test case does not test the recent_syslog() call."""
        glob_mock.return_value = ["/sys/class/net/wlp3s0/wireless"]
        exists_mock.side_effect = [True, False]
        rfkill_list = (
            b"0: hci0: Bluetooth\n	Soft blocked: no\n	Hard blocked: no\n"
            b"1: phy0: Wireless LAN\n	Soft blocked: no\n	Hard blocked: no\n"
        )
        iw_dev_wlp3s0_link_params = (
            b"	freq: 2437.0\n"
            b"	RX: 2363829 bytes (3994 packets)\n"
            b"	TX: 303619 bytes (1560 packets)\n"
            b"	signal: -43 dBm\n"
            b"	rx bitrate: 286.7 MBit/s HE-MCS 11 HE-NSS 2 HE-GI 0 HE-DCM 0\n"
            b"	tx bitrate: 286.7 MBit/s HE-MCS 11 HE-NSS 2 HE-GI 0 HE-DCM 0\n"
            b"	bss flags: CTS-protection short-preamble short-slot-time\n"
            b"	dtim period: 2\n"
            b"	beacon int: 100\n"
        )
        iw_dev_wlp3s0_link = (
            b"Connected to 00:53:e9:15:6c:33 (on wlp3s0)\n	SSID: MyWifiNetwork\n"
            + iw_dev_wlp3s0_link_params
        )
        run_mock.side_effect = [
            subprocess.CompletedProcess(
                args=MagicMock(), returncode=0, stdout=rfkill_list, stderr=b""
            ),
            subprocess.CompletedProcess(
                args=MagicMock(), returncode=0, stdout=iw_dev_wlp3s0_link, stderr=b""
            ),
            subprocess.CompletedProcess(
                args=MagicMock(), returncode=0, stdout=IW_REG_LIST_DE, stderr=b""
            ),
        ]

        report = apport.report.Report()
        apport.hookutils.attach_wifi(report)

        self.assertEqual(report["WifiSyslog"], "some recent logs")
        self.assertEqual(report["RfKill"], rfkill_list.decode("utf-8").rstrip())
        self.assertEqual(
            report["IwDevWlp3s0Link"],
            "Connected to <hidden-mac> (on wlp3s0)\n	SSID: <hidden>\n"
            + iw_dev_wlp3s0_link_params.decode("utf-8").rstrip(),
        )
        self.assertEqual(report["CRDA"], IW_REG_LIST_DE.decode("utf-8").rstrip())
        self.assertEqual(
            set(report.keys()),
            {"CRDA", "Date", "IwDevWlp3s0Link", "ProblemType", "RfKill", "WifiSyslog"},
        )
        glob_mock.assert_called_once_with("/sys/class/net/*/wireless")
        self.assertEqual(
            [call.args[0] for call in exists_mock.call_args_list],
            ["/sbin/iw", "/var/log/wpa_supplicant.log"],
        )
        self.assertEqual(
            [call.args[0] for call in run_mock.call_args_list],
            [["rfkill", "list"], ["iw", "dev", "wlp3s0", "link"], ["iw", "reg", "get"]],
        )

    @unittest.mock.patch("os.path.exists")
    @unittest.mock.patch("subprocess.run")
    @unittest.mock.patch(
        "apport.hookutils.recent_syslog", MagicMock(return_value="some recent logs")
    )
    def test_attach_wifi_without_iw(
        self, run_mock: MagicMock, exists_mock: MagicMock
    ) -> None:
        """Test attach_wifi() with no iw command installed.

        This test case does not test the recent_syslog() call and rfkill command."""
        exists_mock.return_value = False
        run_mock.return_value = subprocess.CompletedProcess(
            args=MagicMock(), returncode=0, stdout=b"rfkill output", stderr=b""
        )

        report = apport.report.Report()
        apport.hookutils.attach_wifi(report)

        self.assertEqual(report["WifiSyslog"], "some recent logs")
        self.assertEqual(report["RfKill"], "rfkill output")
        self.assertEqual(report["CRDA"], "N/A")
        self.assertEqual(
            set(report.keys()), {"CRDA", "Date", "ProblemType", "RfKill", "WifiSyslog"}
        )
        exists_mock.assert_called()
        self.assertEqual(
            [call.args[0] for call in exists_mock.call_args_list],
            ["/sbin/iw", "/var/log/wpa_supplicant.log"],
        )
        run_mock.assert_called_once()
        self.assertEqual(run_mock.call_args_list[0].args[0], ["rfkill", "list"])

    def test_path_to_key(self) -> None:
        """Transform a file path to a valid report key."""
        self.assertEqual(apport.hookutils.path_to_key("simple.txt"), "simple.txt")
        self.assertEqual(
            apport.hookutils.path_to_key("path/with/dirs.txt"), "path.with.dirs.txt"
        )
        self.assertEqual(
            apport.hookutils.path_to_key('/funny:characters!& ".txt'),
            ".funny.characters.._..txt",
        )

    @unittest.mock.patch("subprocess.Popen")
    @unittest.mock.patch("os.path.exists", MagicMock(return_value=True))
    def test_recent_syslog_journald_cmd(self, popen_mock: MagicMock) -> None:
        class _SkipPopen(Exception):
            pass

        popen_mock.side_effect = _SkipPopen

        cmd = ["journalctl", "--quiet", "-b", "-a"]
        cmd_system_only = cmd + ["--system"]

        with self.assertRaises(_SkipPopen):
            apport.hookutils.recent_syslog(re.compile("."))
        popen_mock.assert_called_once_with(cmd_system_only, stdout=unittest.mock.ANY)

        popen_mock.reset_mock()
        with self.assertRaises(_SkipPopen):
            apport.hookutils.recent_syslog(re.compile("."), journald_only_system=True)
        popen_mock.assert_called_once_with(cmd_system_only, stdout=unittest.mock.ANY)

        popen_mock.reset_mock()
        with self.assertRaises(_SkipPopen):
            apport.hookutils.recent_syslog(re.compile("."), journald_only_system=False)
        popen_mock.assert_called_once_with(cmd, stdout=unittest.mock.ANY)

    def test_deprecated_upstart_functions(self) -> None:
        """attach_upstart_*() throw deprecation warnings."""
        with self.assertWarns(PendingDeprecationWarning):
            apport.hookutils.attach_upstart_logs({}, "apport")
        with self.assertWarns(PendingDeprecationWarning):
            apport.hookutils.attach_upstart_overrides({}, "nonexisting")

    def test_recent_syslog_race_condition(self) -> None:
        """Test recent_syslog reads stdout buffer after process completion.

        LP: #2073935
        """
        with unittest.mock.patch("apport.hookutils.subprocess.Popen") as popen_mock:
            process = unittest.mock.Mock()
            popen_mock.return_value.__enter__.return_value = process
            process.returncode = 0

            process.poll = Mock(return_value=0)
            process.stdout = MagicMock()
            process.stdout.__iter__.side_effect = [iter([b"content"])]

            match_content = apport.hookutils.recent_syslog(
                re.compile("."), path="/nonexisting"
            )
            self.assertEqual(match_content, "content")

    def test_recent_syslog_long_process(self) -> None:
        """Test recent_syslog retrieves entire stdout."""
        with unittest.mock.patch("apport.hookutils.subprocess.Popen") as popen_mock:
            process = unittest.mock.Mock()
            popen_mock.return_value.__enter__.return_value = process
            process.returncode = 0

            process.poll = Mock(side_effect=[None, 0])
            process.stdout = MagicMock()
            process.stdout.__iter__.side_effect = [
                iter([b"test\n"]),
                iter([b"content\n"]),
            ]

            match_content = apport.hookutils.recent_syslog(
                re.compile("."), path="/nonexisting"
            )

            self.assertEqual(match_content, "test\ncontent\n")

    @staticmethod
    @unittest.mock.patch("subprocess.run")
    def test_execute_multiple_root_commands_no_commands(run_mock: MagicMock) -> None:
        outputs = apport.hookutils.execute_multiple_root_commands({})
        assert not outputs
        assert not run_mock.called
